#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections

import numpy as np

import montblanc

from hypercube import HyperCube
import pyrap.tables as pt

# Map MS column string types to numpy types
MS_TO_NP_TYPE_MAP = {
    'INT' : np.int32,
    'FLOAT' : np.float32,
    'DOUBLE' : np.float64,
    'BOOLEAN' : np.bool_,
    'COMPLEX' : np.complex64,
    'DCOMPLEX' : np.complex128
}

# Key names for main and taql selected tables
MAIN_TABLE = 'MAIN'
ORDERED_MAIN_TABLE = 'ORDERED_MAIN'
ORDERED_UVW_TABLE = 'ORDERED_UVW'
ORDERED_TIME_TABLE = 'ORDERED_TIME'
ORDERED_BASELINE_TABLE = 'ORDERED_BASELINE'

# Measurement Set sub-table name string constants
ANTENNA_TABLE = 'ANTENNA'
SPECTRAL_WINDOW_TABLE = 'SPECTRAL_WINDOW'
DATA_DESCRIPTION_TABLE = 'DATA_DESCRIPTION'
POLARIZATION_TABLE = 'POLARIZATION'
FIELD_TABLE = 'FIELD'

SUBTABLE_KEYS = (ANTENNA_TABLE,
    SPECTRAL_WINDOW_TABLE,
    DATA_DESCRIPTION_TABLE,
    POLARIZATION_TABLE,
    FIELD_TABLE)

# Main MS column name constants
TIME = 'TIME'
ANTENNA1 = 'ANTENNA1'
ANTENNA2 = 'ANTENNA2'
UVW = 'UVW'
DATA = 'DATA'
FLAG = 'FLAG'
WEIGHT = 'WEIGHT'
MODEL_DATA = 'MODEL_DATA'
CORRECTED_DATA = 'CORRECTED_DATA'

# Antenna sub-table column name constants
POSITION = 'POSITION'

# Field sub-table column name constants
PHASE_DIR = 'PHASE_DIR'

# Spectral window sub-table column name constants
CHAN_FREQ = 'CHAN_FREQ'
NUM_CHAN='NUM_CHAN'
REF_FREQUENCY = 'REF_FREQUENCY'

# Columns used in select statement
SELECTED = [TIME, ANTENNA1, ANTENNA2, UVW,
    DATA, MODEL_DATA, CORRECTED_DATA, FLAG, WEIGHT]

# Named tuple defining a mapping from MS row to dimension
OrderbyMap = collections.namedtuple("OrderbyMap", "dimension orderby")

# Mappings for time, baseline and band
TIME_MAP = OrderbyMap("ntime", "TIME")
BASELINE_MAP = OrderbyMap("nbl", "ANTENNA1, ANTENNA2")
BAND_MAP = OrderbyMap("nbands", "[SELECT SPECTRAL_WINDOW_ID "
        "FROM ::DATA_DESCRIPTION][DATA_DESC_ID]")

# Place mapping in a list
MS_ROW_MAPPINGS = [
    TIME_MAP,
    BASELINE_MAP,
    BAND_MAP
]

UPDATE_DIMENSIONS = ['ntime', 'nbl', 'na', 'nchan', 'nbands', 'npol',
    'npolchan', 'nvis']

# Main measurement set ordering dimensions
MS_DIM_ORDER = ('ntime', 'nbl', 'nbands')
# UVW measurement set ordering dimensions
UVW_DIM_ORDER = ('ntime', 'nbl')


def orderby_clause(dimensions, unique=False):
    columns = ", ".join(m.orderby for m
        in MS_ROW_MAPPINGS if m.dimension in dimensions)

    return " ".join(("ORDERBY", "UNIQUE" if unique else "", columns))

def subtable_name(msname, subtable=None):
    return '::'.join((msname, subtable)) if subtable else msname

def open_table(msname, subtable=None):
    return pt.table(subtable_name(msname, subtable),
        ack=False, readonly=False)

def row_extents(cube, dim_order=None):
    if dim_order is None:
        dim_order = MS_DIM_ORDER

    shape = cube.dim_global_size(*dim_order)
    lower = cube.dim_lower_extent(*dim_order)
    upper = tuple(u-1 for u in cube.dim_upper_extent(*dim_order))

    return (np.ravel_multi_index(lower, shape),
        np.ravel_multi_index(upper, shape) + 1)

def uvw_row_extents(cube):
    return row_extents(cube, UVW_DIM_ORDER)

class MeasurementSetManager(object):
    def __init__(self, msname, slvr_cfg):
        super(MeasurementSetManager, self).__init__()

        self._msname = msname
        # Create dictionary of tables
        self._tables = { k: open_table(msname, k) for k in SUBTABLE_KEYS }

        if not pt.tableexists(msname):
            raise ValueError("'{ms}' does not exist "
                "or is not a Measurement Set!".format(ms=msname))

        # Add imaging columns, just in case
        pt.addImagingColumns(msname, ack=False)

        # Open the main measurement set
        ms = open_table(msname)

        # Access individual tables
        ant, spec, ddesc, pol, field = (self._tables[k] for k in SUBTABLE_KEYS)

        # Sanity check the polarizations
        if pol.nrows() > 1:
            raise ValueError("Multiple polarization configurations!")

        self._npol = npol = pol.getcol('NUM_CORR')[0]

        if npol != 4:
            raise ValueError('Expected four polarizations')

        # Number of channels per band
        chan_per_band = spec.getcol('NUM_CHAN')

        # Require the same number of channels per band
        if not all(chan_per_band[0] == cpb for cpb in chan_per_band):
            raise ValueError('Channels per band {cpb} are not equal!'
                .format(cpb=chan_per_band))

        if ddesc.nrows() != spec.nrows():
            raise ValueError("DATA_DESCRIPTOR.nrows() "
                "!= SPECTRAL_WINDOW.nrows()")

        # Hard code auto-correlations and field_id 0
        self._auto_correlations = auto_correlations = slvr_cfg['auto_correlations']
        self._field_id = field_id = 0

        # Create a view over the MS, ordered by
        # (1) time (TIME)
        # (2) baseline (ANTENNA1, ANTENNA2)
        # (3) band (SPECTRAL_WINDOW_ID via DATA_DESC_ID)
        ordering_query = " ".join((
            "SELECT FROM $ms",
            "WHERE FIELD_ID={fid}".format(fid=field_id),
            "" if auto_correlations else "AND ANTENNA1 != ANTENNA2",
            orderby_clause(MS_DIM_ORDER)
        ))

        # Ordered Measurement Set
        oms = pt.taql(ordering_query)

        montblanc.log.debug("MS ordering query is '{o}'."
            .format(o=ordering_query))

        # Measurement Set ordered by unique time and baseline
        otblms = pt.taql("SELECT FROM $oms {c}".format(
            c=orderby_clause(UVW_DIM_ORDER, unique=True)))

        # Store the main table
        self._tables[MAIN_TABLE] = ms
        self._tables[ORDERED_MAIN_TABLE] = oms
        self._tables[ORDERED_UVW_TABLE] = otblms

        self._column_descriptors = {col: ms.getcoldesc(col) for col in SELECTED}

        # Count distinct timesteps in the MS
        t_orderby = orderby_clause(['ntime'], unique=True)
        t_query = "SELECT FROM $otblms {c}".format(c=t_orderby)
        self._tables[ORDERED_TIME_TABLE] = ot = pt.taql(t_query)
        self._ntime = ntime = ot.nrows()

        # Count number of baselines in the MS
        bl_orderby = orderby_clause(['nbl'], unique=True)
        bl_query = "SELECT FROM $otblms {c}".format(c=bl_orderby)
        self._tables[ORDERED_BASELINE_TABLE] = obl = pt.taql(bl_query)
        self._nbl = nbl = obl.nrows()

        # Number of channels per band
        self._nchanperband = chan_per_band[0]

        self._nchan = nchan = sum(chan_per_band)
        self._nbands = nbands = len(chan_per_band)
        self._npolchan = npolchan = npol*nchan
        self._nvis = nvis = ntime*nbl*nchan

        # Update the cube with dimension information
        # obtained from the MS
        updated_sizes = [ntime, nbl, ant.nrows(),
            sum(chan_per_band), len(chan_per_band), npol,
            npolchan, nvis]

        self._dim_sizes = dim_sizes = { dim: size for dim, size
            in zip(UPDATE_DIMENSIONS, updated_sizes) }

        shape = tuple(dim_sizes[d] for d in MS_DIM_ORDER)
        expected_rows = np.product(shape)

        if not expected_rows == oms.nrows():
            dim_desc = ", ".join('(%s,%s)' % (d, s) for
                d, s in zip(MS_DIM_ORDER, shape))
            row_desc = " x ".join('%s' % s for s in shape)

            montblanc.log.warning("Encountered '{msr}' rows in '{ms}' "
                "but expected '{rd} = {er}' after finding the following "
                "dimensions by inspection: [{d}]. Irregular Measurement Sets "
                "are not fully supported due to the generality of the format.".format(
                    msr=oms.nrows(), ms=msname,
                    er=expected_rows, rd=row_desc, d=dim_desc))

    def close(self):
        # Close all the tables
        for table in list(self._tables.values()):
            table.close()

    @property
    def msname(self):
        return self._msname

    @property
    def column_descriptors(self):
        return self._column_descriptors

    @property
    def channels_per_band(self):
        return self._nchanperband

    def updated_dimensions(self):
        return [(k, v) for k, v in list(self._dim_sizes.items())]

    @property
    def auto_correlations(self):
        return self._auto_correlations

    @property
    def field_id(self):
        return self._field_id

    @property
    def main_table(self):
        return self._tables[MAIN_TABLE]

    @property
    def ordered_main_table(self):
        return self._tables[ORDERED_MAIN_TABLE]

    @property
    def ordered_uvw_table(self):
        return self._tables[ORDERED_UVW_TABLE]

    @property
    def ordered_time_table(self):
        return self._tables[ORDERED_TIME_TABLE]

    @property
    def antenna_table(self):
        return self._tables[ANTENNA_TABLE]

    @property
    def spectral_window_table(self):
        return self._tables[SPECTRAL_WINDOW_TABLE]

    @property
    def data_description_table(self):
        return self._tables[DATA_DESCRIPTION_TABLE]

    @property
    def polarization_table(self):
        return self._tables[POLARIZATION_TABLE]

    @property
    def field_table(self):
        return self._tables[FIELD_TABLE]

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etraceback):
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import numpy as np

import math
import re

import montblanc

from .parallactic_angles import parallactic_angles
from .ant_uvw import antenna_uvw

from montblanc.src_types import (
    source_types,
    source_nr_vars,
    default_sources,
    sources_to_nr_vars,
    source_range,
    source_range_tuple,
    source_range_slices)


from .parsing import parse_python_assigns

def nr_of_baselines(na, auto_correlations=False):
    """
    Compute the number of baselines for the
    given number of antenna. Can specify whether
    auto-correlations should be taken into
    account
    """
    m = (na-1) if auto_correlations is False else (na+1)
    return (na*m)//2

def nr_of_antenna(nbl, auto_correlations=False):
    """
    Compute the number of antenna for the
    given number of baselines. Can specify whether
    auto-correlations should be taken into
    account
    """
    t = 1 if auto_correlations is False else -1
    return int(t + math.sqrt(1 + 8*nbl)) // 2

def blocks_required(N, threads_per_block):
    """
    Returns the number of blocks required, given
    N, the total number of threads, and threads_per_block
    """
    return (N + threads_per_block - 1) / threads_per_block

def fmt_bytes(nbytes):
    """ Returns a human readable string, given the number of bytes """
    for x in ['B','KB','MB','GB']:
        if nbytes < 1024.0:
            return "%3.1f%s" % (nbytes, x)
        nbytes /= 1024.0

    return "%.1f%s" % (nbytes, 'TB')

def array_bytes(shape, dtype):
    """ Estimates the memory in bytes required for an array of the supplied shape and dtype """
    return np.product(shape)*np.dtype(dtype).itemsize

def random_float(shape, dtype):
    return np.random.random(size=shape).astype(dtype)

def random_complex(shape, ctype):
    if ctype == np.complex64:
        dtype = np.float32
    elif ctype == np.complex128:
        dtype = np.float64
    else:
        raise TypeError("Invalid complex type {ct}".format(ct=ctype))

    return random_float(shape, dtype) + 1j*random_float(shape, dtype)

def random_like(ary=None, shape=None, dtype=None):
    """
    Returns a random array of the same shape and type as the
    supplied array argument, or the supplied shape and dtype
    """
    if ary is not None:
        shape, dtype = ary.shape, ary.dtype
    elif shape is None or dtype is None:
        raise ValueError((
            'random_like(ary, shape, dtype) must be supplied '
            'with either an array argument, or the shape and dtype '
            'of the desired random array.'))

    if np.issubdtype(dtype, np.complexfloating):
        return (np.random.random(size=shape) + \
            np.random.random(size=shape)*1j).astype(dtype)
    else:
        return np.random.random(size=shape).astype(dtype)

def flatten(nested):
    """ Return a flatten version of the nested argument """
    flat_return = list()

    def __inner_flat(nested,flat):
        for i in nested:
            __inner_flat(i, flat) if isinstance(i, list) else flat.append(i)
        return flat

    __inner_flat(nested,flat_return)

    return flat_return

def dict_array_bytes(ary, template):
    """
    Return the number of bytes required by an array

    Arguments
    ---------------
    ary : dict
        Dictionary representation of an array
    template : dict
        A dictionary of key-values, used to replace any
        string values in the array with concrete integral
        values

    Returns
    -----------
    The number of bytes required to represent
    the array.
    """
    shape = shape_from_str_tuple(ary['shape'], template)
    dtype = dtype_from_str(ary['dtype'], template)

    return array_bytes(shape, dtype)

def dict_array_bytes_required(arrays, template):
    """
    Return the number of bytes required by
    a dictionary of arrays.

    Arguments
    ---------------
    arrays : list
        A list of dictionaries defining the arrays
    template : dict
        A dictionary of key-values, used to replace any
        string values in the arrays with concrete integral
        values

    Returns
    -----------
    The number of bytes required to represent
    all the arrays.
    """
    return np.sum([dict_array_bytes(ary, template)
        for ary in arrays])

__DIM_REDUCTION_RE = re.compile(    # Capture Groups and Subgroups
    "^\s*(?P<name>[A-Za-z0-9_]*?)"  # 1.   Dimension name
    "(?:\s*?=\s*?"                  # 2.   White spaces and =
        "(?P<value>[0-9]*?)"        # 2.1  A value
        "(?P<percent>\%?)"          # 2.2  Possibly followed by a percentage
    ")?\s*?$")                      #      Capture group 2 possibly occurs

def viable_dim_config(bytes_available, arrays, template,
        dim_ord, nsolvers=1):
    """
    Returns the number of timesteps possible, given the registered arrays
    and a memory budget defined by bytes_available

    Arguments
    ----------------
    bytes_available : int
        The memory budget, or available number of bytes
        for solving the problem.
    arrays : list
        List of dictionaries describing the arrays
    template : dict
        Dictionary containing key-values that will be used
        to replace any string representations of dimensions
        and types. slvr.template_dict() will return something
        suitable.
    dim_ord : list
        list of dimension string names that the problem should be
        subdivided by. e.g. ['ntime', 'nbl', 'nchan'].
        Multple dimensions can be reduced simultaneously using
        the following syntax 'nbl&na'. This is mostly useful for
        the baseline-antenna equivalence.
    nsolvers : int
        Number of solvers to budget for. Defaults to one.

    Returns
    ----------
    A tuple (boolean, dict). The boolean is True if the problem
    can fit within the supplied budget, False otherwise.
    THe dictionary contains the reduced dimensions as key and
    the reduced size as value.
    e.g. (True, { 'time' : 1, 'nbl' : 1 })

    For a dim_ord = ['ntime', 'nbl', 'nchan'], this method will try and fit
    a ntime x nbl x nchan problem into the available number of bytes.
    If this is not possible, it will first set ntime=1, and then try fit an
    1 x nbl x nchan problem into the budget, then a 1 x 1 x nchan
    problem.

    One can specify reductions for specific dimensions.
    For e.g. ['ntime=20', 'nbl=1&na=2', 'nchan=50%']

    will reduce ntime to 20, but no lower. nbl=1&na=2 sets
    both nbl and na to 1 and 2 in the same operation respectively.
    nchan=50\% will continuously halve the nchan dimension
    until it reaches a value of 1.
    """

    if not isinstance(dim_ord, list):
        raise TypeError('dim_ord should be a list')

    # Don't accept non-negative memory budgets
    if bytes_available < 0:
        bytes_available = 0

    modified_dims = {}
    T = template.copy()

    bytes_used = dict_array_bytes_required(arrays, T)*nsolvers

    # While more bytes are used than are available, set
    # dimensions to one in the order specified by the
    # dim_ord argument.
    while bytes_used > bytes_available:
        try:
            dims = dim_ord.pop(0)
            montblanc.log.debug('Applying reduction {s}. '
                'Bytes available: {a} used: {u}'.format(
                    s=dims,
                    a=fmt_bytes(bytes_available),
                    u=fmt_bytes(bytes_used)))
            dims = dims.strip().split('&')
        except IndexError:
            # No more dimensions available for reducing
            # the problem size. Unable to fit the problem
            # within the specified memory budget
            return False, modified_dims

        # Can't fit everything into memory,
        # Lower dimensions and re-evaluate
        for dim in dims:
            match = re.match(__DIM_REDUCTION_RE, dim)

            if not match:
                raise ValueError(
                    "{d} is an invalid dimension reduction string "
                    "Valid strings are for e.g. "
                    "'ntime', 'ntime=20' or 'ntime=20%'"
                        .format(d=dim))

            dim_name = match.group('name')
            dim_value = match.group('value')
            dim_percent = match.group('percent')
            dim_value = 1 if dim_value is None else int(dim_value)

            # Attempt reduction by a percentage
            if dim_percent == '%':
                dim_value = int(T[dim_name] * int(dim_value) / 100.0)
                if dim_value < 1:
                    # This can't be reduced any further
                    dim_value = 1
                else:
                    # Allows another attempt at reduction
                    # by percentage on this dimension
                    dim_ord.insert(0, dim)

            # Apply the dimension reduction
            if T[dim_name] > dim_value:
                modified_dims[dim_name] = dim_value
                T[dim_name] = dim_value
            else:
                montblanc.log.info(('Ignored reduction of {d} '
                    'of size {s} to {v}. ').format(
                        d=dim_name, s=T[dim_name], v=dim_value))

        bytes_used = dict_array_bytes_required(arrays, T)*nsolvers

    return True, modified_dims

def dtype_from_str(sdtype, template):
    """
    Substitutes string dtype parameters
    using a a property dictionary

    Parameters
    ----------
        sdtype :
            string defining the dtype. e.g. 'ft'
        template:
            Dictionary contain actual types, {'ft' : np.float32' }

    Returns
        sdtype if it isn't a string
        template[sdtype] otherwise

    """

    if not isinstance(sdtype, str):
        return sdtype

    return template[sdtype]

def shape_from_str_tuple(sshape, variables, ignore=None):
    """
    Substitutes string values in the supplied shape parameter
    with integer variables stored in a dictionary

    Parameters
    ----------
    sshape : tuple/string composed of integers and strings.
        The strings should related to integral properties
        registered with this Solver object
    variables : dictionary
        Keys with associated integer values. Used to replace
        string values within the tuple
    ignore : list
        A list of tuple strings to ignore

    >>> print self.shape_from_str_tuple((4,'na','ntime'),ignore=['ntime'])
    (4, 3)
    """
    if ignore is None: ignore = []

    if not isinstance(sshape, tuple) and not isinstance(sshape, list):
        raise TypeError('sshape argument must be a tuple or list')

    if not isinstance(ignore, list):
        raise TypeError('ignore argument must be a list')

    return tuple([int(eval_expr(v,variables)) if isinstance(v,str) else int(v)
        for v in sshape if v not in ignore])

def shape_list(l,shape,dtype):
    """ Shape a list of lists into the appropriate shape and data type """
    return np.array(l, dtype=dtype).reshape(shape)

def array_convert_function(sshape_one, sshape_two, variables):
    """ Return a function defining the conversion process between two NumPy
    arrays of different shapes """
    if not isinstance(sshape_one, tuple): sshape_one = (sshape_one,)
    if not isinstance(sshape_two, tuple): sshape_two = (sshape_two,)

    s_one = flatten([eval_expr_names_and_nrs(d) if isinstance(d,str) else d
        for d in sshape_one])
    s_two = flatten([eval_expr_names_and_nrs(d) if isinstance(d,str) else d
        for d in sshape_two])

    if len(s_one) != len(s_two):
        raise ValueError(('Flattened shapes %s and %s '\
            'do not have the same length. '
            'Original shapes were %s and %s') % \
            (s_one, s_two, sshape_one, sshape_two))

    # Reason about the transpose
    t_idx = tuple([s_one.index(v) for v in s_two])

    # Figure out the actual numeric shape values to use
    n_one = shape_from_str_tuple(s_one, variables)
    n_two = [eval_expr(d,variables)
        if isinstance(d,str) else d for d in sshape_two]

    def f(ary): return np.reshape(ary, n_one).transpose(t_idx).reshape(n_two)

    return f

def redistribute_threads(blockdimx, blockdimy, blockdimz,
    dimx, dimy, dimz):
    """
    Redistribute threads from the Z dimension towards the X dimension.
    Also clamp number of threads to the problem dimension size,
    if necessary
    """

    # Shift threads from the z dimension
    # into the y dimension
    while blockdimz > dimz:
        tmp = blockdimz // 2
        if tmp < dimz:
            break

        blockdimy *= 2
        blockdimz = tmp

    # Shift threads from the y dimension
    # into the x dimension
    while blockdimy > dimy:
        tmp = blockdimy // 2
        if tmp < dimy:
            break

        blockdimx *= 2
        blockdimy = tmp

    # Clamp the block dimensions
    # if necessary
    if dimx < blockdimx:
        blockdimx = dimx

    if dimy < blockdimy:
        blockdimy = dimy

    if dimz < blockdimz:
        blockdimz = dimz

    return blockdimx, blockdimy, blockdimz


def register_default_dimensions(cube, slvr_cfg):
    """ Register the default dimensions for a RIME solver """

    import montblanc.src_types as mbs

    # Pull out the configuration options for the basics
    autocor = slvr_cfg['auto_correlations']

    ntime = 10
    na = 7
    nbands = 1
    nchan = 16
    npol = 4

    # Infer number of baselines from number of antenna,
    nbl = nr_of_baselines(na, autocor)

    if not npol == 4:
        raise ValueError("npol set to {}, but only 4 polarisations "
                         "are currently supported.")

    # Register these dimensions on this solver.
    cube.register_dimension('ntime', ntime,
        description="Timesteps")
    cube.register_dimension('na', na,
        description="Antenna")
    cube.register_dimension('nbands', nbands,
        description="Bands")
    cube.register_dimension('nchan', nchan,
        description="Channels")
    cube.register_dimension('npol', npol,
        description="Polarisations")
    cube.register_dimension('nbl', nbl,
        description="Baselines")

    # Register dependent dimensions
    cube.register_dimension('npolchan', nchan*npol,
        description='Polarised channels')
    cube.register_dimension('nvis', ntime*nbl*nchan,
        description='Visibilities')

    # Convert the source types, and their numbers
    # to their number variables and numbers
    # { 'point':10 } => { 'npsrc':10 }
    src_cfg = default_sources()
    src_nr_vars = sources_to_nr_vars(src_cfg)
    # Sum to get the total number of sources
    cube.register_dimension('nsrc', sum(src_nr_vars.values()),
        description="Sources (Total)")

    # Register the individual source types
    for nr_var, nr_of_src in list(src_nr_vars.items()):
        cube.register_dimension(nr_var, nr_of_src,
            description='{} sources'.format(mbs.SOURCE_DIM_TYPES[nr_var]))

class ContextWrapper(object):
    """ Context Manager Wrapper for CUDA Contexts! """
    def __init__(self, context):
        self.context = context

    def __enter__(self):
        """ Pushed the wrapped context onto the stack """
        self.context.push()
        return self

    def __exit__(self,type,value,traceback):
        """ Pop when we're done """
        self.context.pop()



try:
    from future_builtins import zip
except:
    pass

from itertools import islice
import math
import numpy as np
from numba import jit

# Coordinate indexing constants
u, v, w = list(range(3))

try:
    isclose = math.isclose
except AttributeError:
    @jit(nopython=True, nogil=True, cache=True)
    def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
        return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


@jit(nopython=True, nogil=True, cache=True)
def _antenna_uvw_loop(uvw, antenna1, antenna2, ant_uvw,
                      chunk_index, start, end):

    c = chunk_index

    # Cluster (first antenna) associated with each antenna
    clusters = np.full((ant_uvw.shape[1],), -1, dtype=antenna1.dtype)

    # Iterate over rows in chunk
    for row in range(start, end):
        a1 = antenna1[row]
        a2 = antenna2[row]

        # Have they been clustered yet?
        cl1 = clusters[a1]
        cl2 = clusters[a2]

        # Both new -- start a new cluster relative to a1
        if cl1 == -1 and cl2 == -1:
            clusters[a1] = clusters[a2] = a1
            ant_uvw[c, a1, u] = 0.0
            ant_uvw[c, a1, v] = 0.0
            ant_uvw[c, a1, w] = 0.0

            # If this is not an auto-correlation
            # assign a2 to inverse of baseline UVW,
            if a1 != a2:
                ant_uvw[c, a2, u] = uvw[row, u]
                ant_uvw[c, a2, v] = uvw[row, v]
                ant_uvw[c, a2, w] = uvw[row, w]

        # if either antenna has not been clustered,
        # infer its coordinate from the clustered one
        elif cl1 != -1 and cl2 == -1:
            clusters[a2] = cl1
            ant_uvw[c, a2, u] = ant_uvw[c, a1, u] + uvw[row, u]
            ant_uvw[c, a2, v] = ant_uvw[c, a1, v] + uvw[row, v]
            ant_uvw[c, a2, w] = ant_uvw[c, a1, w] + uvw[row, w]
        elif cl1 == -1 and cl2 != -1:
            clusters[a1] = cl2
            ant_uvw[c, a1, u] = ant_uvw[c, a2, u] - uvw[row, u]
            ant_uvw[c, a1, v] = ant_uvw[c, a2, v] - uvw[row, v]
            ant_uvw[c, a1, w] = ant_uvw[c, a2, w] - uvw[row, w]
        # Both clustered. If clusters differ, merge them
        elif cl1 != -1 and cl2 != -1:
            if cl1 != cl2:
                # how much do we need to add to the current cluster2
                # reference position to make the baseline a2 - a1 consistent?
                u_off = uvw[row, u] - (ant_uvw[c, a2, u] - ant_uvw[c, a1, u])
                v_off = uvw[row, v] - (ant_uvw[c, a2, v] - ant_uvw[c, a1, v])
                w_off = uvw[row, w] - (ant_uvw[c, a2, w] - ant_uvw[c, a1, w])

                # Merge cluster2 into cluster1
                for ant, cluster in enumerate(clusters):
                    if cluster == cl2:
                        clusters[ant] = cl1
                        ant_uvw[c, ant, u] += u_off
                        ant_uvw[c, ant, v] += v_off
                        ant_uvw[c, ant, w] += w_off

        # Shouldn't ever occur
        else:
            raise ValueError("Illegal Condition")


@jit(nopython=True, nogil=True, cache=True)
def _antenna_uvw(uvw, antenna1, antenna2, chunks, nr_of_antenna):
    """ numba implementation of antenna_uvw """

    if antenna1.ndim != 1:
        raise ValueError("antenna1 shape should be (row,)")

    if antenna2.ndim != 1:
        raise ValueError("antenna2 shape should be (row,)")

    if uvw.ndim != 2 or uvw.shape[1] != 3:
        raise ValueError("uvw shape should be (row, 3)")

    if not (uvw.shape[0] == antenna1.shape[0] == antenna2.shape[0]):
        raise ValueError("First dimension of uvw, antenna1 "
                         "and antenna2 do not match")

    if chunks.ndim != 1:
        raise ValueError("chunks shape should be (utime,)")

    if nr_of_antenna < 1:
        raise ValueError("nr_of_antenna < 1")

    ant_uvw_shape = (chunks.shape[0], nr_of_antenna, 3)
    antenna_uvw = np.full(ant_uvw_shape, np.nan, dtype=uvw.dtype)

    start = 0

    for ci, chunk in enumerate(chunks):
        end = start + chunk

        # one pass should be enough!
        _antenna_uvw_loop(uvw, antenna1, antenna2, antenna_uvw, ci, start, end)

        start = end

    return antenna_uvw


class AntennaUVWDecompositionError(Exception):
    pass


def _raise_decomposition_errors(uvw, antenna1, antenna2,
                                chunks, ant_uvw, max_err):
    """ Raises informative exception for an invalid decomposition """

    start = 0

    problem_str = []

    for ci, chunk in enumerate(chunks):
        end = start + chunk

        ant1 = antenna1[start:end]
        ant2 = antenna2[start:end]
        cuvw = uvw[start:end]

        ant1_uvw = ant_uvw[ci, ant1, :]
        ant2_uvw = ant_uvw[ci, ant2, :]
        ruvw = ant2_uvw - ant1_uvw

        # Identifty rows where any of the UVW components differed
        close = np.isclose(ruvw, cuvw)
        problems = np.nonzero(np.logical_or.reduce(np.invert(close), axis=1))

        for row in problems[0]:
            problem_str.append("[row %d [%d, %d] (chunk %d)]: "
                               "original %s recovered %s "
                               "ant1 %s ant2 %s" % (
                                    start+row, ant1[row], ant2[row], ci,
                                    cuvw[row], ruvw[row],
                                    ant1_uvw[row], ant2_uvw[row]))

            # Exit inner loop early
            if len(problem_str) >= max_err:
                break

        # Exit outer loop early
        if len(problem_str) >= max_err:
            break

        start = end

    # Return early if nothing was wrong
    if len(problem_str) == 0:
        return

    # Add a preamble and raise exception
    problem_str = ["Antenna UVW Decomposition Failed",
                   "The following differences were found "
                   "(first 100):"] + problem_str
    raise AntennaUVWDecompositionError('\n'.join(problem_str))


class AntennaMissingError(Exception):
    pass


def _raise_missing_antenna_errors(ant_uvw, max_err):
    """ Raises an informative error for missing antenna """

    # Find antenna uvw coordinates where any UVW component was nan
    # nan + real == nan
    problems = np.nonzero(np.add.reduce(np.isnan(ant_uvw), axis=2))
    problem_str = []

    for c, a in zip(*problems):
        problem_str.append("[chunk %d antenna %d]" % (c, a))

        # Exit early
        if len(problem_str) >= max_err:
            break

    # Return early if nothing was wrong
    if len(problem_str) == 0:
        return

    # Add a preamble and raise exception
    problem_str = ["Antenna were missing"] + problem_str
    raise AntennaMissingError('\n'.join(problem_str))


def antenna_uvw(uvw, antenna1, antenna2, chunks,
                nr_of_antenna, check_missing=False,
                check_decomposition=False, max_err=100):
    """
    Computes per-antenna UVW coordinates from baseline ``uvw``,
    ``antenna1`` and ``antenna2`` coordinates logically grouped
    into baseline chunks.

    The example below illustrates two baseline chunks
    of size 6 and 5, respectively.

    .. code-block:: python

        uvw = ...
        ant1 = np.array([0, 0, 0, 1, 1, 2, 0, 0, 0, 1, 1], dtype=np.int32)
        ant2 = np.array([1, 2, 3, 2, 3, 3, 1, 2, 3, 1, 2], dtype=np.int32)
        chunks = np.array([6, 5], dtype=np.int32)

        ant_uv = antenna_uvw(uvw, ant1, ant2, chunks, nr_of_antenna=4)

    The first antenna of the first baseline of a chunk is chosen as the origin
    of the antenna coordinate system, while the second antenna is set to the
    negative of the baseline UVW coordinate. Subsequent antenna UVW coordinates
    are iteratively derived from the first two coordinates. Thus,
    the baseline indices need not be properly ordered (within the chunk).

    If it is not possible to derive coordinates for an antenna,
    it's coordinate will be set to nan.

    Parameters
    ----------
    uvw : np.ndarray
        Baseline UVW coordinates of shape (row, 3)
    antenna1 : np.ndarray
        Baseline first antenna of shape (row,)
    antenna2 : np.ndarray
        Baseline second antenna of shape (row,)
    chunks : np.ndarray
        Number of baselines per unique timestep with shape (chunks,)
        :code:`np.sum(chunks) == row` should hold.
    nr_of_antenna : int
        Total number of antenna in the solution.
    check_missing (optional) : bool
        If ``True`` raises an exception if it was not possible
        to compute UVW coordinates for all antenna (i.e. some were nan).
        Defaults to ``False``.
    check_decomposition (optional) : bool
        If ``True``, checks that the antenna decomposition accurately
        reproduces the coordinates in ``uvw``, or that
        :code:`ant_uvw[c,ant1,:] - ant_uvw[c,ant2,:] == uvw[s:e,:]`
        where ``s`` and ``e`` are the start and end rows
        of chunk ``c`` respectively. Defaults to ``False``.
    max_err (optional) : integer
        Maximum numbers of errors when checking for missing antenna
        or innacurate decompositions. Defaults to ``100``.

    Returns
    -------
    np.ndarray
        Antenna UVW coordinates of shape (chunks, nr_of_antenna, 3)
    """

    ant_uvw = _antenna_uvw(uvw, antenna1, antenna2, chunks, nr_of_antenna)

    if check_missing:
        _raise_missing_antenna_errors(ant_uvw, max_err=max_err)

    if check_decomposition:
        _raise_decomposition_errors(uvw, antenna1, antenna2, chunks,
                                    ant_uvw, max_err=max_err)

    return ant_uvw



import os
import unittest
import timeit

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib

lightspeed = 299792458.


def complex_phase(lm, uvw, frequency):
    """
    Compute the complex phase from lm, uvw and frequency expressions
    """

    # Get the dynamic shape of input tensors
    lm_shape = tf.shape(input=lm)
    uvw_shape = tf.shape(input=uvw)
    frequency_shape = tf.shape(input=frequency)

    # The shapes are themselves tensors
    nsrc = lm_shape[0]
    ntime, na = uvw_shape[0], uvw_shape[1]
    nchan = frequency_shape[0]

    # Define some constants
    one = tf.constant(1.0, dtype=dtype)
    minus_two_pi_over_C = tf.constant(-2.0*np.pi/lightspeed, dtype=dtype)

    # Reshape now so that we get broadcasting in later operations
    # Need to pack list since list contains tensors, e.g. nsrc
    l = tf.reshape(lm[:, 0], tf.stack([nsrc, 1, 1, 1]))
    m = tf.reshape(lm[:, 1], tf.stack([nsrc, 1, 1, 1]))

    u = tf.reshape(uvw[:, :, 0], tf.stack([1, ntime, na, 1]))
    v = tf.reshape(uvw[:, :, 1], tf.stack([1, ntime, na, 1]))
    w = tf.reshape(uvw[:, :, 2], tf.stack([1, ntime, na, 1]))

    frequency = tf.reshape(frequency, tf.stack([1, 1, 1, nchan]))

    n = tf.sqrt(one - l**2 - m**2) - one

    # Outer product l*u + m*v * n*w
    phase = tf.convert_to_tensor(value=l*u + m*v + n*w, name='real_phase')

    # Multiply in constants
    phase = minus_two_pi_over_C*phase*frequency

    # No GPU implementation of exp yet
    #return tf.exp(tf.complex(0.0, phase), name='complex_phase')
    return tf.complex(tf.cos(phase), tf.sin(phase))


def complex_phase_numpy(lm, uvw, frequency):
    nsrc, _ = lm.shape
    ntime, na, _ = uvw.shape
    nchan, = frequency.shape

    lm = lm.reshape(nsrc, 1, 1, 1, 2)
    uvw = uvw.reshape(1, ntime, na, 1, 3)
    frequency = frequency.reshape(1, 1, 1, nchan)

    l, m = lm[:, :, :, :, 0], lm[:, :, :, :, 1]
    u, v, w = uvw[:, :, :, :, 0], uvw[:, :, :, :, 1], uvw[:, :, :, :, 2]

    n = np.sqrt(1.0 - l**2 - m**2) - 1.0
    real_phase = -2*np.pi*1j*(l*u + m*v + n*w)*frequency/lightspeed
    return np.exp(real_phase)


class TestComplexPhase(unittest.TestCase):
    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_complex_phase(self):
        type_permutations = [
            [np.float32, np.complex64],
            [np.float64, np.complex128]]

        for FT, CT in type_permutations:
            self._impl_test_complex_phase(FT, CT)

    def _impl_test_complex_phase(self, FT, CT):
        nsrc, ntime, na, nchan = 100, 50, 64, 128

        # Set up our numpy input arrays
        np_lm = np.random.random(size=(nsrc, 2)).astype(FT)*0.1
        np_uvw = np.random.random(size=(ntime, na, 3)).astype(FT)
        np_frequency = np.linspace(1.3e9, 1.5e9, nchan,
                                   endpoint=True, dtype=FT)

        np_args = [np_lm, np_uvw, np_frequency]
        tf_names = ['lm', 'uvw', 'frequency']

        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, tf_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.phase(*tf_args, CT=CT)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        # Now create a tensorflow Session to evaluate the above
        with tf.compat.v1.Session() as S:
            S.run(init_op)

            phase_np = complex_phase_numpy(np_lm, np_uvw, np_frequency)
            phase_cpu = S.run(cpu_op)
            assert np.allclose(phase_cpu, phase_np)

            for phase_gpu in S.run(gpu_ops):
                assert np.allclose(phase_cpu, phase_gpu)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from ..context_help import context_help
from ..hypercube_proxy_metaclass import HypercubeProxyMetaClass
import six

class _setter_property(object):
    def __init__(self, func, doc=None):
        self.func = func
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        return self.func(obj, value)

@six.add_metaclass(HypercubeProxyMetaClass)
class SourceContext(object):
    """
    Context object passed to data sources.

    It provides information to the user implementing a data source
    about the extents of the data tile that should be provided.

    .. code-block:: python

        # uvw varies by time and baseline and has 3 coordinate components
        cube.register_array("uvw", ("ntime", "nbl", 3), np.float64)

        ...

        class UVWSourceProvider(SourceProvider):
            def __init__(self, uvw_data):
                # All UVW coordinates
                self._uvw_data = uvw_data

            def uvw(self, context):
                print context.help(display_cube=True)

                # Query dimensions directly
                (lt, ut), (lb, ub) = context.dim_extents("ntime", "nbl")
                # Get the cube extents, ignoring
                # last dimension which is always (0, 3)
                (lt, ut), (lb, ub), (_, _) = context.array_extents("uvw")
                # Return data tile from larger array
                return self._uvw_data[lt:ut, lb:ub, :]


    Public methods of a :py:class:`~hypercube.base_cube.HyperCube`
    are proxied on this object. Other useful information, such
    as the configuration, iteration space arguments, expected
    array shape and data type, and the abstract array schema
    are also present on this object.


    """
    __slots__ = ('_cube', '_cfg', '_name', '_shape', '_dtype',
        '_iter_args', '_array_schema')

    def __init__(self, name, cube, slvr_cfg, iter_args,
                array_schema, shape, dtype):
        self._name = name
        self._cube = cube
        self._cfg = slvr_cfg
        self._iter_args = iter_args
        self._array_schema = array_schema
        self._shape = shape
        self._dtype = dtype

    @_setter_property
    def cube(self, value):
        self._cube = value

    @property
    def cfg(self):
        """
        Configuration
        """
        return self._cfg

    @cfg.setter
    def cfg(self, value):
        self._cfg = value

    @property
    def shape(self):
        """
        The expected shape of the array
        that should be produced by the data source
        """
        return self._shape

    @shape.setter
    def shape(self, value):
        self._shape = value

    @property
    def dtype(self):
        """
        The expected data type of the array
        that should be produced by the data source
        """
        return self._dtype

    @dtype.setter
    def dtype(self, value):
        self._dtype = value

    @property
    def name(self):
        """ The name of the data source of this context. """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def array_schema(self):
        """
        The array schema of the array associated
        with this data source. For instance if `model_vis` is
        registered on a hypercube as follows:

        .. code-block:: python

            # Register model_vis array_schema on hypercube
            cube.register_array("model_vis",
                ("ntime", "nbl", "nchan", "ncorr"),
                np.complex128)

            ...
            # Create a source context for model_vis data source
            context = SourceContext("model_vis", ...)
            ...
            # Obtain the array schema
            context.array_schema == ("ntime", "nbl", "nchan", "ncorr")

        """


        return self._array_schema

    @property
    def iter_args(self):
        """
        Iteration arguments that describe the tile sizes
        over which iteration is performed. In the following example,
        iteration is occuring in tiles of 100 Timesteps, 64 Channels
        and 50 Point Sources.

        .. code-block:: python

            context.iter_args == [("ntime", 100),
                    ("nchan", 64), ("npsrc", 50)]
        """
        return self._iter_args

    def help(self, display_cube=False):
        """
        Get help associated with this context

        Args
        -----
        display_cube: bool
            Add hypercube description to the output
        Returns
        -------
            str
                A help string associated with this context
        """
        return context_help(self, display_cube)





import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

from tensorflow.python.client import device_lib




class TestSumCoherencies(unittest.TestCase):
    """ Tests the SumCoherencies operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()

        # Load the custom operation library
        #self.rime = tf.load_op_library('rime.so')
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_sum_coherencies(self):
        """ Test the SumCoherencies operator """

        # List of type constraints for testing this operator
        type_permutations = [
            [np.float32, np.complex64, {'rtol': 1e-4}],
            [np.float64, np.complex128, {}]]

        # Run test with the type combinations above
        for FT, CT, cmp_kwargs in type_permutations:
            self._impl_test_sum_coherencies(FT, CT, cmp_kwargs)

    def _impl_test_sum_coherencies(self, FT, CT, cmp_kwargs):
        """ Implementation of the SumCoherencies operator test """

        def rf(*a, **kw):
            return np.random.random(*a, **kw).astype(FT)

        def rc(*a, **kw):
            return rf(*a, **kw) + 1j*rf(*a, **kw).astype(CT)

        nsrc, ntime, na, nchan = 10, 15, 7, 16
        nbl = na*(na-1)//2

        np_ant1, np_ant2 = [np.int32(x) for x in np.triu_indices(na, 1)]
        np_ant1, np_ant2 = (np.tile(np_ant1, ntime).reshape(ntime, nbl),
                            np.tile(np_ant2, ntime).reshape(ntime, nbl))
        np_shape = rf(size=(nsrc, ntime, nbl, nchan))
        np_ant_jones = rc(size=(nsrc, ntime, na, nchan, 4))
        np_sgn_brightness = np.random.randint(0, 3, size=(nsrc, ntime, nchan), dtype=np.int8) - 1
        np_base_coherencies = rc(size=(ntime, nbl, nchan, 4))

        # Argument list
        np_args = [np_ant1, np_ant2, np_shape, np_ant_jones,
                   np_sgn_brightness, np_base_coherencies]
        # Argument string name list
        arg_names = ['antenna1', 'antenna2', 'shape', 'ant_jones',
                     'sgn_brightness', 'base_coherencies']
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.sum_coherencies(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU coherencies
            cpu_coherencies = S.run(cpu_op)

            #Compare against the GPU coherencies
            for gpu_coherencies in S.run(gpu_ops):
               self.assertTrue(np.allclose(cpu_coherencies, gpu_coherencies,
                               **cmp_kwargs))


if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


import collections
import functools
import types

import numpy as np

import montblanc.util as mbu
import montblanc.impl.rime.tensorflow.ms.ms_manager as MS

from montblanc.impl.rime.tensorflow.sources.source_provider import SourceProvider
from montblanc.impl.rime.tensorflow.sources import SourceContext

class MSSourceProvider(SourceProvider):
    """
    Source Provider that retrieves input data from a
    MeasurementSet
    """

    def __init__(self, manager, vis_column=None):
        """
        Constructs an MSSourceProvider object

        Parameters
        ----------
        manager: :py:class:`.MeasurementSetManager`
            The :py:class:`.MeasurementSetManager` used to access
            the Measurement Set.
        vis_column: str
            Column from which observed visibilities will be read
        """
        self._manager = manager
        self._name = "Measurement Set '{ms}'".format(ms=manager.msname)

        self._vis_column = 'DATA' if vis_column is None else vis_column

        # Cache columns on the object
        # Handle these columns slightly differently
        # They're used to compute the parallactic angle
        # TODO: Fit them into the cache_ms_read strategy at some point

        # Cache antenna positions
        self._antenna_positions = manager.antenna_table.getcol(MS.POSITION)

        # Cache timesteps
        self._times = manager.ordered_time_table.getcol(MS.TIME)

        # Cache the phase direction for the field
        # [0][0] because (a) we select only 1 row
        #                (b) assumes a NUM_POLY of 1
        self._phase_dir = manager.field_table.getcol(MS.PHASE_DIR,
            startrow=manager.field_id, nrow=1)[0][0]

    def name(self):
        return self._name

    def updated_dimensions(self):
        # Defer to manager's method
        return self._manager.updated_dimensions()

    def phase_centre(self, context):
        return self._phase_dir.astype(context.dtype)

    def antenna_position(self, context):
        la, ua = context.dim_extents('na')
        return (self._antenna_positions[la:ua]
                    .astype(context.dtype))

    def time(self, context):
        lt, ut = context.dim_extents('ntime')
        return self._times[lt:ut].astype(context.dtype)

    def frequency(self, context):
        """ Frequency data source """
        channels = self._manager.spectral_window_table.getcol(MS.CHAN_FREQ)
        return channels.reshape(context.shape).astype(context.dtype)

    def ref_frequency(self, context):
        """ Reference frequency data source """
        num_chans = self._manager.spectral_window_table.getcol(MS.NUM_CHAN)
        ref_freqs = self._manager.spectral_window_table.getcol(MS.REF_FREQUENCY)

        data = np.hstack((np.repeat(rf, bs) for bs, rf in zip(num_chans, ref_freqs)))
        return data.reshape(context.shape).astype(context.dtype)

    def uvw(self, context):
        """ Per-antenna UVW coordinate data source """

        # Hacky access of private member
        cube = context._cube

        # Create antenna1 source context
        a1_actual = cube.array("antenna1", reify=True)
        a1_ctx = SourceContext("antenna1", cube, context.cfg,
            context.iter_args, cube.array("antenna1"),
            a1_actual.shape, a1_actual.dtype)

        # Create antenna2 source context
        a2_actual = cube.array("antenna2", reify=True)
        a2_ctx = SourceContext("antenna2", cube, context.cfg,
            context.iter_args, cube.array("antenna2"),
            a2_actual.shape, a2_actual.dtype)

        # Get antenna1 and antenna2 data
        ant1 = self.antenna1(a1_ctx).ravel()
        ant2 = self.antenna2(a2_ctx).ravel()

        # Obtain per baseline UVW data
        lrow, urow = MS.uvw_row_extents(context)
        uvw = self._manager.ordered_uvw_table.getcol(MS.UVW,
                                                startrow=lrow,
                                                nrow=urow-lrow)

        # Perform the per-antenna UVW decomposition
        ntime, nbl = context.dim_extent_size('ntime', 'nbl')
        na = context.dim_global_size('na')
        chunks = np.repeat(nbl, ntime).astype(ant1.dtype)

        auvw = mbu.antenna_uvw(uvw, ant1, ant2, chunks, nr_of_antenna=na)

        return auvw.reshape(context.shape).astype(context.dtype)

    def antenna1(self, context):
        """ antenna1 data source """
        lrow, urow = MS.uvw_row_extents(context)
        antenna1 = self._manager.ordered_uvw_table.getcol(
            MS.ANTENNA1, startrow=lrow, nrow=urow-lrow)

        return antenna1.reshape(context.shape).astype(context.dtype)

    def antenna2(self, context):
        """ antenna2 data source """
        lrow, urow = MS.uvw_row_extents(context)
        antenna2 = self._manager.ordered_uvw_table.getcol(
            MS.ANTENNA2, startrow=lrow, nrow=urow-lrow)

        return antenna2.reshape(context.shape).astype(context.dtype)

    def parallactic_angles(self, context):
        """ parallactic angle data source """
        # Time and antenna extents
        (lt, ut), (la, ua) = context.dim_extents('ntime', 'na')

        return (mbu.parallactic_angles(self._times[lt:ut],
                self._antenna_positions[la:ua], self._phase_dir)
                                            .reshape(context.shape)
                                            .astype(context.dtype))


    def observed_vis(self, context):
        """ Observed visibility data source """
        lrow, urow = MS.row_extents(context)

        data = self._manager.ordered_main_table.getcol(
            self._vis_column, startrow=lrow, nrow=urow-lrow)

        return data.reshape(context.shape).astype(context.dtype)

    def flag(self, context):
        """ Flag data source """
        lrow, urow = MS.row_extents(context)

        flag = self._manager.ordered_main_table.getcol(
            MS.FLAG, startrow=lrow, nrow=urow-lrow)

        return flag.reshape(context.shape).astype(context.dtype)

    def weight(self, context):
        """ Weight data source """
        lrow, urow = MS.row_extents(context)

        weight = self._manager.ordered_main_table.getcol(
            MS.WEIGHT, startrow=lrow, nrow=urow-lrow)

        # WEIGHT is applied across all channels
        weight = np.repeat(weight, self._manager.channels_per_band, 0)
        return weight.reshape(context.shape).astype(context.dtype)

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etraceback):
        self.close()

    def __str__(self):
        return self.__class__.__name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

def load_tf_lib():
    """ Load the tensorflow library """
    from os.path import join as pjoin
    from pkg_resources import working_set
    from pkg_resources import Requirement

    import tensorflow as tf
    import os

    path = pjoin('ext', 'rime.so')
    mbloc = pjoin(working_set.find(Requirement.parse('montblanc')).location, "montblanc")
    rime_lib_path = pjoin(mbloc, path)
    if not os.path.isfile(rime_lib_path):
        from montblanc import ext
        rime_lib_path = os.path.join(os.path.dirname(ext.__file__), 'rime.so')
    if not os.path.isfile(rime_lib_path):
        raise RuntimeError(f"Montblanc backend not found: '{rime_lib_path}'. Have you compiled the backend?")
    return tf.load_op_library(rime_lib_path)




from functools import partial
import textwrap

from ruamel.yaml import YAML

def load_config(file):
    yaml = YAML()
    return yaml.load(file)

def raise_validator_errors(validator):
    """
    Raise any errors associated with the validator.

    Parameters
    ----------
    validator : :class:`cerberus.Validator`
        Validator

    Raises
    ------
    ValueError
        Raised if errors existed on `validator`.
        Message describing each error and information
        associated with the configuration option
        causing the error.
    """

    if len(validator._errors) == 0:
        return

    def _path_str(path, name=None):
        """ String of the document/schema path. `cfg["foo"]["bar"]` """
        L = [name] if name is not None else []
        L.extend('["%s"]' % p for p in path)
        return "".join(L)

    def _path_leaf(path, dicts):
        """ Dictionary Leaf of the schema/document given the path """
        for p in path:
            dicts = dicts[p]

        return dicts

    wrap = partial(textwrap.wrap, initial_indent=' '*4,
                                subsequent_indent=' '*8)

    msg = ["There were configuration errors:"]

    for e in validator._errors:
        schema_leaf = _path_leaf(e.document_path, validator.schema)
        doc_str = _path_str(e.document_path, "cfg")

        msg.append("Invalid configuration option %s == '%s'." % (doc_str, e.value))

        try:
            otype = schema_leaf["type"]
            msg.extend(wrap("Type must be '%s'." % otype))
        except KeyError:
            pass

        try:
            allowed = schema_leaf["allowed"]
            msg.extend(wrap("Allowed values are '%s'." % allowed))
        except KeyError:
            pass

        try:
            description = schema_leaf["__description__"]
            msg.extend(wrap("Description: %s" % description))
        except KeyError:
            pass

    raise ValueError("\n".join(msg))


def config_validator():
    from cerberus import Validator
    from montblanc.src_types import default_sources

    class DescriptionValidator(Validator):
        def _validate___description__(self, __description__,
                                                field, value):
            """
            Dummy rule so that '__description__' keys
            can be placed in the schema

            The rule's arguments are validated against this schema:
            {'type': 'string'}
            """
            pass

    schema = {
        'data_source': {
            'type': 'string',
            'allowed': ['default', 'test'],
            'default': 'default',
            '__description__': "Data Source for initialising inputs. "
                               "If 'default', initialised with defaults, "
                               "If 'test', initialised with sensible "
                               "test data." },

        'device_type': {
            'type': 'string',
            'allowed': ['CPU', 'GPU'],
            'default': 'GPU',
            '__description__': "Default compute device." },

        'dtype': {
            'type': 'string',
            'allowed': ['float', 'double'],
            'default': 'double',
            '__description__': "Floating Point precision of "
                                "inputs and solutions." },

        'auto_correlations': {
            'type': 'boolean',
            'default': False,
            '__description__': "Take auto-correlations into account "
                               "when computing number of baselines "
                               "from number of antenna." },

        'polarisation_type': {
            'type': 'string',
            'allowed': ['linear', 'circular'],
            'default': 'linear',
            '__description__': "Type of polarisation. "
                               "Should be 'linear' or 'circular'." },

        'mem_budget': {
            'type': 'integer',
            'min': 1024,
            'default': 1024*1024*1024,
            '__description__': "Memory budget for solving a single "
                               "tile of the problem on a CPU/GPU "
                               "in bytes." },

        'source_batch_size': {
            'type': 'integer',
            'min': 0,
            'default': 500 },

        'version': {
            'type': 'string',
            'default': 'tf' },
    }

    return DescriptionValidator(schema)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import inspect

class AbstractSourceProvider(object):

    def name(self):
        """ Return the name associated with this data source """
        raise NotImplementedError()

    def init(self, init_context):
        """ Called when initialising Providers """
        raise NotImplementedError()

    def start(self, start_context):
        """ Called at the start of any solution """
        raise NotImplementedError()

    def stop(self, stop_context):
        """ Called at the end of any solution """
        raise NotImplementedError()

    def close(self):
        """ Perform any required cleanup """
        raise NotImplementedError()

    def sources(self):
        """ Returns a dictionary of source methods, keyed on source name """
        raise NotImplementedError()

    def updated_dimensions(self):
        """ Return an iterable/mapping of hypercube dimensions to update """
        raise NotImplementedError()

    def updated_arrays(self):
        """ Return an iterable/mapping of hypercube arrays to update """
        raise NotImplementedError()

DEFAULT_ARGSPEC = ['self', 'context']

def find_sources(obj, argspec=None):
    """
    Returns a dictionary of source methods found on this object,
    keyed on method name. Source methods are identified.by argspec,
    a list of argument specifiers. So for e.g. an argpsec of
    :code:`[['self', 'context'], ['s', 'c']]` would match
    methods looking like:

    .. code-block:: python

        def f(self, context):
        ...

    .. code-block:: python

        def f(s, c):
        ...

    is but not

    .. code-block:: python

        def f(self, ctx):
        ...


    """

    if argspec is None:
        argspec = [DEFAULT_ARGSPEC]

    return { n: m for n, m in inspect.getmembers(obj, callable)
        if not n.startswith('_') and
        inspect.getfullargspec(m).args in argspec }


class SourceProvider(AbstractSourceProvider):

    def init(self, init_context):
        """ Called when initialising Providers """
        pass

    def start(self, start_context):
        """ Called at the start of any solution """
        pass

    def stop(self, stop_context):
        """ Called at the end of any solution """
        pass

    def close(self):
        """ Perform any required cleanup. """
        pass

    def sources(self):
        """
        Returns a dictionary of source methods found on this object,
        keyed on method name. Source methods are identified by
        (self, context) arguments on this object. For example:

        .. code-block:: python

            def f(self, context):
                    ...

        is a source method, but

        .. code-block:: python

            def f(self, ctx):
                ...

            is not.

        """

        try:
            return self._sources
        except AttributeError:
            self._sources = find_sources(self)

        return self._sources

    def updated_dimensions(self):
        """ Return an iterable/mapping of hypercube dimensions to update """
        return ()

    def updated_arrays(self):
        """ Return an iterable/mapping of hypercube arrays to update """
        return ()

    def __str__(self):
        return self.name()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

import montblanc

from montblanc.impl.rime.tensorflow.sinks.sink_provider import SinkProvider
import montblanc.impl.rime.tensorflow.ms.ms_manager as MS

class MSSinkProvider(SinkProvider):
    """
    Sink Provider that receives model visibilities produced by
    montblanc
    """

    def __init__(self, manager, vis_column=None):
        """
        Constructs an MSSinkProvider object

        Parameters
        ----------
        manager: :py:class:`.MeasurementSetManager`
            The :py:class:`.MeasurementSetManager` used to access
            the Measurement Set.
        vis_column: str
            Column to which model visibilities will be read
        """

        self._manager = manager
        self._name = "Measurement Set '{ms}'".format(ms=manager.msname)
        self._vis_column = ('CORRECTED_DATA' if vis_column is None else vis_column)

    def name(self):
        return self._name

    def model_vis(self, context):
        """ model visibility data sink """
        column = self._vis_column
        msshape = None

        # Do we have a column descriptor for the supplied column?
        try:
            coldesc = self._manager.column_descriptors[column]
        except KeyError as e:
            coldesc = None

        # Try to get the shape from the descriptor
        if coldesc is not None:
            try:
                msshape = [-1] + coldesc['shape'].tolist()
            except KeyError as e:
                msshape = None

        # Otherwise guess it and warn
        if msshape is None:
            guessed_shape = [self._manager._nchan, 4]

            montblanc.log.warning("Could not obtain 'shape' from the '{c}' "
                "column descriptor. Guessing it is '{gs}'.".format(
                    c=column, gs=guessed_shape))

            msshape = [-1] + guessed_shape

        lrow, urow = MS.row_extents(context)

        self._manager.ordered_main_table.putcol(column,
            context.data.reshape(msshape),
            startrow=lrow, nrow=urow-lrow)

    def __str__(self):
        return self.__class__.__name__




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections
import functools
import os
import re
import string
import sys
import types

import numpy as np
from astropy.io import fits

from hypercube import HyperCube

import montblanc
import montblanc.util as mbu
from montblanc.impl.rime.tensorflow.sources.source_provider import SourceProvider

class FitsFilenameTemplate(string.Template):
    """
    Overrides the ${identifer} braced pattern in the string Template
    with a $(identifier) braced pattern expected by FITS beam filename
    schema
    """
    pattern = r"""
    %(delim)s(?:
      (?P<escaped>%(delim)s)   |   # Escape sequence of two delimiters
      (?P<named>%(id)s)        |   # delimiter and a Python identifier
      \((?P<braced>%(id)s)\)   |   # delimiter and a braced identifier
      (?P<invalid>)                # Other ill-formed delimiter exprs
    )
    """ % { 'delim' : re.escape(string.Template.delimiter),
        'id' : string.Template.idpattern }

class FitsAxes(object):
    """
    FitsAxes object, inspired by Tigger's FITSAxes
    """
    def __init__(self, header):
        self._ndims = ndims = header['NAXIS']

        # Extract header information for each dimension
        axr = list(range(1, ndims+1))
        self._naxis = [header.get('NAXIS%d'%n)      for n in axr]
        self._ctype = [header.get('CTYPE%d'%n, n)   for n in axr]
        self._crval = [header.get('CRVAL%d'%n, 0)   for n in axr]
        self._crpix = [header.get('CRPIX%d'%n)-1    for n in axr]
        self._cdelta = [header.get('CDELT%d'%n, 1)  for n in axr]
        self._cunit = [header.get('CUNIT%d'%n, '').strip().upper()
            for n in axr]

        # Check for custom irregular grid format.
        # Currently only implemented for FREQ dimension.
        irregular_grid = [[header.get('G%s%d' % (self._ctype[i], j), None)
            for j in range(1, self._naxis[i]+1)]
            for i in range(ndims)]

        # Irregular grids are only valid if values exist for all grid points
        valid = [all(x is not None for x in irregular_grid[i])
            for i in range(ndims)]

        def _regular_grid(a, i):
            """ Construct a regular grid from a FitsAxes object and index """
            R = np.arange(0.0, float(a.naxis[i]))
            return (R - a.crpix[i])*a.cdelta[i] + a.crval[i]

        # Set up the grid
        self._grid = [_regular_grid(self, i) if not valid[i]
            else np.asarray(irregular_grid[i]) for i in range(ndims)]

        # Copy original CRVAL and CRDELTA in case they are scaled
        self._scale = [1.0 for n in axr]
        self._crval0 = [v for v in self._crval]
        self._cdelta0 = [v for v in self._cdelta]

        # Map axis names to integers
        self._iaxis = {n: i for i, n in enumerate(self._ctype)}

    @property
    def ndims(self):
        return self._ndims

    def iaxis(self, name):
        try:
            return self._iaxis[name]
        except KeyError:
            return -1

    @property
    def grid(self):
        return self._grid

    @property
    def crpix(self):
        return self._crpix

    @property
    def naxis(self):
        return self._naxis

    @property
    def crval(self):
        return self._crval

    @property
    def cdelta(self):
        return self._cdelta

    @property
    def crval0(self):
        return self._crval0

    @property
    def cdelta0(self):
        return self._cdelta0

    @property
    def cunit(self):
        return self._cunit

    @property
    def ctype(self):
        return self._ctype

    @property
    def scale(self):
        return self._scale

    @property
    def extents(self):
        f = lambda v, i: (v - self.crpix[i])*self.cdelta[i] + self.crval[i]
        return [tuple(f(v, i)  for v in (0, self.naxis[i]-1) )
            for i in range(self.ndims)]

    def set_axis_scale(self, index, scale):
        self.scale[index] = scale
        self.crval[index] = self.crval0[index]*scale
        self.cdelta[index] = self.cdelta0[index]*scale

CIRCULAR_CORRELATIONS = ('rr', 'rl', 'lr', 'll')
LINEAR_CORRELATIONS = ('xx', 'xy', 'yx', 'yy')
REIM = ('re', 'im')

def _create_filenames(filename_schema, feed_type):
    """
    Returns a dictionary of beam filename pairs,
    keyed on correlation,from the cartesian product
    of correlations and real, imaginary pairs

    Given 'beam_$(corr)_$(reim).fits' returns:
    {
      'xx' : ('beam_xx_re.fits', 'beam_xx_im.fits'),
      'xy' : ('beam_xy_re.fits', 'beam_xy_im.fits'),
      ...
      'yy' : ('beam_yy_re.fits', 'beam_yy_im.fits'),
    }

    Given 'beam_$(CORR)_$(REIM).fits' returns:
    {
      'xx' : ('beam_XX_RE.fits', 'beam_XX_IM.fits'),
      'xy' : ('beam_XY_RE.fits', 'beam_XY_IM.fits'),
      ...
      'yy' : ('beam_YY_RE.fits', 'beam_YY_IM.fits'),
    }

    """
    template = FitsFilenameTemplate(filename_schema)

    def _re_im_filenames(corr, template):
        try:
            return tuple(template.substitute(
                corr=corr.lower(), CORR=corr.upper(),
                reim=ri.lower(), REIM=ri.upper())
                    for ri in REIM)
        except KeyError:
            raise ValueError("Invalid filename schema '%s'. "
                            "FITS Beam filename schemas "
                            "must follow forms such as "
                            "'beam_$(corr)_$(reim).fits' or "
                            "'beam_$(CORR)_$(REIM).fits." % filename_schema)

    if feed_type == 'linear':
        CORRELATIONS = LINEAR_CORRELATIONS
    elif feed_type == 'circular':
        CORRELATIONS = CIRCULAR_CORRELATIONS
    else:
        raise ValueError("Invalid feed_type '{}'. "
            "Should be 'linear' or 'circular'")

    return collections.OrderedDict(
        (c, _re_im_filenames(c, template))
        for c in CORRELATIONS)

def _open_fits_files(filenames):
    """
    Given a {correlation: filename} mapping for filenames
    returns a {correlation: file handle} mapping
    """
    kw = { 'mode' : 'update', 'memmap' : False }

    def _fh(fn):
        """ Returns a filehandle or None if file does not exist """
        return fits.open(fn, **kw) if os.path.exists(fn) else None

    return collections.OrderedDict(
            (corr, tuple(_fh(fn) for fn in files))
        for corr, files in list(filenames.items()) )

def _cube_extents(axes, l_ax, m_ax, f_ax, l_sign, m_sign):
    # List of (lower, upper) extent tuples for the given dimensions
    it = list(zip((l_ax, m_ax, f_ax), (l_sign, m_sign, 1.0)))
    # Get the extents, flipping the sign on either end if required
    extent_list = [tuple(s*e for e in axes.extents[i]) for i, s in it]

    # Return [[l_low, u_low, f_low], [l_high, u_high, f_high]]
    return np.array(extent_list).T

def _create_axes(filenames, file_dict):
    """ Create a FitsAxes object """

    try:
        # Loop through the file_dictionary, finding the
        # first open FITS file.
        f = next(iter(f for tup in list(file_dict.values())
            for f in tup if f is not None))
    except StopIteration as e:
        raise ValueError("No FITS files were found. "
            "Searched filenames: '{f}'." .format(
                f=list(filenames.values())))


    # Create a FitsAxes object
    axes = FitsAxes(f[0].header)

    # Scale any axes in degrees to radians
    for i, u in enumerate(axes.cunit):
        if u == 'DEG':
            axes.cunit[i] = 'RAD'
            axes.set_axis_scale(i, np.pi/180.0)

    return axes

def _axis_and_sign(ax_str):
    """ Extract axis and sign from given axis string """
    return (ax_str[1:], -1.0) if ax_str[0] == '-' else (ax_str, 1.0)

class FitsBeamSourceProvider(SourceProvider):
    """
    Feeds holography cubes from a series of eight FITS files matching a
    filename_schema. A schema of :code:`'beam_$(corr)_$(reim).fits'`
    matches:

    .. code-block:: python

        ['beam_xx_re.fits', 'beam_xx_im.fits',
         'beam_xy_re.fits', 'beam_xy_im.fits',
          ...
          'beam_yy_re.fits', 'beam_yy_im.fits']

    while :code:`'beam_$(CORR)_$(REIM).fits'` matches

    .. code-block:: python

        ['beam_XX_RE.fits', 'beam_XX_IM.fits',
          'beam_XY_RE.fits', 'beam_XY_IM.fits',
          ...
          'beam_YY_RE.fits', 'beam_YY_IM.fits']


    Missing files will result in zero values for that correlation
    and real/imaginary component. The shape of the FITS data will be
    inferred from the first file found and subsequent files should match
    that shape.

    The type of correlation will be derived from the feed type.
    Currently, linear :code:`['xx', 'xy', 'yx', 'yy']` and
    circular :code:`['rr', 'rl', 'lr', 'll']` are supported.
    """
    def __init__(self, filename_schema, l_axis=None, m_axis=None):
        """
        Constructs a FitsBeamSourceProvider object

        Parameters
        ----------
            filename_schema : str
                See :py:class:`.FitsBeamSourceProvider` for valid schemas
            l_axis : str
                FITS axis interpreted as the L axis. `L` and `X` are
                sensible values here. `-L` will invert the coordinate
                system on that axis.
            m_axis : str
                FITS axis interpreted as the M axis. `M` and `Y` are
                sensible values here. `-M` will invert the coordinate
                system on that axis.
        """
        l_axis, l_sign = _axis_and_sign('L' if l_axis is None else l_axis)
        m_axis, m_sign = _axis_and_sign('M' if m_axis is None else m_axis)

        self._l_axis = l_axis
        self._l_sign = l_sign
        self._m_axis = m_axis
        self._m_sign = m_sign

        self._fits_dims = fits_dims = (l_axis, m_axis, 'FREQ')
        self._beam_dims = ('beam_lw', 'beam_mh', 'beam_nud')

        self._filename_schema = filename_schema
        self._name = "FITS Beams '{s}'".format(s=filename_schema)

        # Have we initialised this object?
        self._initialised = False

    def _initialise(self, feed_type="linear"):
        """
        Initialise the object by generating appropriate filenames,
        opening associated file handles and inspecting the FITS axes
        of these files.
        """
        self._filenames = filenames = _create_filenames(self._filename_schema,
                                                        feed_type)
        self._files = files = _open_fits_files(filenames)
        self._axes = axes = _create_axes(filenames, files)
        self._dim_indices = dim_indices = l_ax, m_ax, f_ax = tuple(
            axes.iaxis(d) for d in self._fits_dims)

        # Complain if we can't find required axes
        for i, ax in zip(dim_indices, self._fits_dims):
            if i == -1:
                raise ValueError("'%s' axis not found!" % ax)

        self._cube_extents = _cube_extents(axes, l_ax, m_ax, f_ax,
            self._l_sign, self._m_sign)
        self._shape = tuple(axes.naxis[d] for d in dim_indices) + (4,)
        self._beam_freq_map = axes.grid[f_ax]

        # Now describe our dimension sizes
        self._dim_updates = [(n, axes.naxis[i]) for n, i
            in zip(self._beam_dims, dim_indices)]

        self._initialised = True

    def name(self):
        """ Name of this Source Provider """
        return self._name

    def init(self, init_context):
        """ Perform any initialisation """
        self._initialise(init_context.cfg['polarisation_type'])

    def ebeam(self, context):
        """ ebeam cube data source """
        if context.shape != self.shape:
            raise ValueError("Partial feeding of the "
                "beam cube is not yet supported %s %s." % (context.shape, self.shape))

        ebeam = np.empty(context.shape, context.dtype)

        # Iterate through the correlations,
        # assigning real and imaginary data, if present,
        # otherwise zeroing the correlation
        for i, (re, im) in enumerate(self._files.values()):
            ebeam[:,:,:,i].real[:] = 0 if re is None else re[0].data.T
            ebeam[:,:,:,i].imag[:] = 0 if im is None else im[0].data.T

        return ebeam

    def beam_extents(self, context):
        """ Beam extent data source """
        return self._cube_extents.flatten().astype(context.dtype)

    def beam_freq_map(self, context):
        """ Beam frequency map data source """
        return self._beam_freq_map.astype(context.dtype)

    def updated_dimensions(self):
        """ Indicate dimension sizes """
        return self._dim_updates

    @property
    def filename_schema(self):
        """ Filename schema """
        return self._filename_schema

    @property
    def shape(self):
        """ Shape of the beam cube """
        return self._shape

    def close(self):
        if not hasattr(self, "_files"):
            return

        for re, im in list(self._files.values()):
            re.close()
            im.close()

        self._files.clear()

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etrace):
        self.close()

    def __str__(self):
        return self.__class__.__name__


import argparse
import re

from .op_source_templates import (MAIN_HEADER_TEMPLATE,
    CPP_HEADER_TEMPLATE,
    CPP_SOURCE_TEMPLATE,
    CUDA_HEADER_TEMPLATE,
    CUDA_SOURCE_TEMPLATE,
    PYTHON_SOURCE_TEMPLATE)

MODULE='rime'
PROJECT='montblanc'

# Convert CamelCase op names to snake_case
def camel_to_snake_case(name):
    FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
    ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')
    s1 = FIRST_CAP_RE.sub(r'\1_\2', name)
    return ALL_CAP_RE.sub(r'\1_\2', s1).lower()

# Derive a C++ header guard from the header name
def header_guard(header_name):
    guard_str = header_name.replace('.', '_')
    return ''.join([MODULE, '_', guard_str]).upper()

parser = argparse.ArgumentParser()
parser.add_argument('opname')
args = parser.parse_args()

snake_case = camel_to_snake_case(args.opname)

# Create dictionary with variables required for creating the templates
D = {
    'opname' : args.opname,
    'project' : PROJECT,
    'module' : MODULE,
    'snake_case' : snake_case,
    'library' : ''.join([MODULE, '.so']),
}

# Filenames
D.update({
    'main_header_file' : ''.join([snake_case, '_op.h']),
    'cpp_header_file' : ''.join([snake_case, '_op_cpu.h']),
    'cpp_source_file' : ''.join([snake_case, '_op_cpu.cpp']),
    'cuda_header_file' : ''.join([snake_case, '_op_gpu.cuh']),
    'cuda_source_file' : ''.join([snake_case, '_op_gpu.cu']),
    'python_test_file' : ''.join(['test_', snake_case, '.py'])
})

# C++ header guards
D.update({
    'main_header_guard' : header_guard(D['main_header_file']),
    'cpp_header_guard' : header_guard(D['cpp_header_file']),
    'cuda_header_guard' : header_guard(D['cuda_header_file']),
})

# C++ namespace
D.update({
    'project_namespace_start' : ''.join([PROJECT, '_namespace_begin']).upper(),
    'project_namespace_stop' : ''.join([PROJECT, '_namespace_stop']).upper(),
    'op_namespace_start' : ''.join([PROJECT, '_', snake_case, '_namespace_begin']).upper(),
    'op_namespace_stop' : ''.join([PROJECT, '_', snake_case, '_namespace_stop']).upper(),
})

# kernel names
D.update({
    'kernel_name' : ''.join([MODULE, '_', snake_case])
})

# Write out each file, substituting template variables
with open(D['main_header_file'], 'w') as f:
    f.write(MAIN_HEADER_TEMPLATE.substitute(**D))

with open(D['cpp_header_file'], 'w') as f:
    f.write(CPP_HEADER_TEMPLATE.substitute(**D))

with open(D['cpp_source_file'], 'w') as f:
    f.write(CPP_SOURCE_TEMPLATE.substitute(**D))

with open(D['cuda_header_file'], 'w') as f:
    f.write(CUDA_HEADER_TEMPLATE.substitute(**D))

with open(D['cuda_source_file'], 'w') as f:
    f.write(CUDA_SOURCE_TEMPLATE.substitute(**D))

with open(D['python_test_file'], 'w') as f:
    f.write(PYTHON_SOURCE_TEMPLATE.substitute(**D))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

import numpy as np

import montblanc

# List of source types and the variable names
# referring to the number of sources for that type
POINT_TYPE = 'point'
POINT_NR_VAR = 'npsrc'

GAUSSIAN_TYPE = 'gaussian'
GAUSSIAN_NR_VAR = 'ngsrc'

SERSIC_TYPE = 'sersic'
SERSIC_NR_VAR = 'nssrc'

# Type to numbering variable mapping,
# with a specific ordering (point first)
SOURCE_VAR_TYPES = OrderedDict([
    (POINT_TYPE, POINT_NR_VAR),
    (GAUSSIAN_TYPE, GAUSSIAN_NR_VAR),
    (SERSIC_TYPE, SERSIC_NR_VAR)
])

SOURCE_DIM_TYPES = OrderedDict((v, k) for (k, v)
    in list(SOURCE_VAR_TYPES.items()))

def source_types():
    """ Returns a list of registered source types """
    return list(SOURCE_VAR_TYPES.keys())

def source_nr_vars():
    """ Returns a list of registered source number variables """
    return list(SOURCE_VAR_TYPES.values())

def source_var_types():
    """ Returns a mapping of source type to number variable """
    return SOURCE_VAR_TYPES

def default_sources(**kwargs):
    """
    Returns a dictionary mapping source types
    to number of sources. If the number of sources
    for the source type is supplied in the kwargs
    these will be placed in the dictionary.

    e.g. if we have 'point', 'gaussian' and 'sersic'
    source types, then

    default_sources(point=10, gaussian=20)

    will return an OrderedDict {'point': 10, 'gaussian': 20, 'sersic': 0}
    """
    S = OrderedDict()
    total = 0

    invalid_types = [t for t in list(kwargs.keys()) if t not in SOURCE_VAR_TYPES]

    for t in invalid_types:
        montblanc.log.warning('Source type %s is not yet '
            'implemented in montblanc. '
            'Valid source types are %s' % (t, list(SOURCE_VAR_TYPES.keys())))

    # Zero all source types
    for k, v in list(SOURCE_VAR_TYPES.items()):
        # Try get the number of sources for this source
        # from the kwargs
        value = kwargs.get(k, 0)

        try:
            value = int(value)
        except ValueError:
            raise TypeError(('Supplied value %s '
                'for source %s cannot be '
                'converted to an integer') % \
                    (value, k))

        total += value
        S[k] = value

    # Add a point source if no others exist
    if total == 0:
        S[POINT_TYPE] = 1

    return S

def sources_to_nr_vars(sources):
    """
    Converts a source type to number of sources mapping into
    a source numbering variable to number of sources mapping.

    If, for example, we have 'point', 'gaussian' and 'sersic'
    source types, then passing the following dict as an argument

    sources_to_nr_vars({'point':10, 'gaussian': 20})

    will return an OrderedDict

    {'npsrc': 10, 'ngsrc': 20, 'nssrc': 0 }
    """

    sources = default_sources(**sources)

    try:
        return OrderedDict((SOURCE_VAR_TYPES[name], nr)
            for name, nr in list(sources.items()))
    except KeyError as e:
        raise KeyError((
            'No source type ''%s'' is '
            'registered. Valid source types '
            'are %s') % (e, list(SOURCE_VAR_TYPES.keys())))

def source_range_tuple(start, end, nr_var_dict):
    """
    Given a range of source numbers, as well as a dictionary
    containing the numbers of each source, returns a dictionary
    containing tuples of the start and end index
    for each source variable type.
    """

    starts = np.array([0 for nr_var in list(SOURCE_VAR_TYPES.values())])
    ends = np.array([nr_var_dict[nr_var] if nr_var in nr_var_dict else 0
        for nr_var in list(SOURCE_VAR_TYPES.values())])
    sum_counts = np.cumsum(ends)
    idx = np.arange(len(starts))

    # Find the intervals containing the
    # start and ending indices
    start_idx, end_idx = np.searchsorted(
        sum_counts, [start, end], side='right')

    # Handle edge cases
    if end >= sum_counts[-1]:
        end = sum_counts[-1]
        end_idx = len(sum_counts) - 1

    # Find out which variable counts fall within the range
    # of the supplied indices and zero those outside this range
    invalid = np.logical_not(np.logical_and(start_idx <= idx, idx <= end_idx))
    starts[invalid] = ends[invalid] = 0

    # Modify the associated starting and ending positions
    starts[start_idx] = start
    ends[end_idx] = end

    if start >= sum_counts[0]:
        starts[start_idx] -= sum_counts[start_idx-1]

    if end >= sum_counts[0]:
        ends[end_idx] -= sum_counts[end_idx-1]

    return OrderedDict((n, (starts[i], ends[i]))
        for i, n in enumerate(SOURCE_VAR_TYPES.values()))

def source_range(start, end, nr_var_dict):
    """
    Given a range of source numbers, as well as a dictionary
    containing the numbers of each source, returns a dictionary
    containing tuples of the start and end index
    for each source variable type.
    """

    return OrderedDict((k, e-s)
        for k, (s, e)
        in list(source_range_tuple(start, end, nr_var_dict).items()))

def source_range_slices(start, end, nr_var_dict):
    """
    Given a range of source numbers, as well as a dictionary
    containing the numbers of each source, returns a dictionary
    containing slices for each source variable type.
    """

    return OrderedDict((k, slice(s,e,1))
        for k, (s, e)
        in list(source_range_tuple(start, end, nr_var_dict).items()))


import unittest
from pprint import pformat

from six.moves import range
import numpy as np

from montblanc.util import antenna_uvw


class TestAntennaUvWDecomposition(unittest.TestCase):
    def test_uvw_antenna(self):
        na = 17
        ntime = 1

        # For both auto correlations and without them
        for auto_cor in (0, 1):
            # Compute default antenna pairs
            ant1, ant2 = np.triu_indices(na, auto_cor)

            # Create random per-antenna UVW coordinates.
            # zeroing the first antenna
            ant_uvw = np.random.random(size=(ntime, na, 3)).astype(np.float64)
            ant_uvw[0, 0, :] = 0

            time_chunks = np.array([ant1.size], dtype=ant1.dtype)

            # Compute per-baseline UVW coordinates.
            bl_uvw = (ant_uvw[:, ant1, :] - ant_uvw[:, ant2, :]).reshape(-1, 3)

            # Now recover the per-antenna and per-baseline UVW coordinates.
            rant_uvw = antenna_uvw(bl_uvw, ant1, ant2, time_chunks,
                                   nr_of_antenna=na, check_decomposition=True)

    def test_uvw_disjoint(self):

        # Three initially disjoint baselines here, but the last baseline [2, 9]
        # connects the first and the last
        # Set 1: 0, 1, 2, 3
        # Set 2: 4, 5, 6, 7, 8
        # Set 3: 8, 10, 11, 12
        # Connection between Set 1 and Set 3 is the last baseline [2, 9]
        ant1 = np.array([1,  2,  3,  4,  5,  5,  7,  9, 10, 11,  2])
        ant2 = np.array([2,  2,  0,  5,  5,  6,  8, 10, 11, 12,  9])

        na = np.unique(np.concatenate([ant1, ant2])).size
        ntime = 1

        # Create random per-antenna UVW coordinates.
        # zeroing the first antenna
        ant_uvw = np.random.random(size=(ntime, na, 3)).astype(np.float64)
        ant_uvw[0, 0, :] = 0

        time_chunks = np.array([ant1.size], dtype=ant1.dtype)

        # Compute per-baseline UVW coordinates.
        bl_uvw = (ant_uvw[:, ant1, :] - ant_uvw[:, ant2, :]).reshape(-1, 3)

        # Now recover the per-antenna and per-baseline UVW coordinates.
        rant_uvw = antenna_uvw(bl_uvw, ant1, ant2, time_chunks,
                               nr_of_antenna=na, check_decomposition=True)

    def test_uvw_antenna_missing_bl_impl(self):
        na = 17
        removed_ants_per_time = ([0, 1, 7], [2, 10, 15, 9], [3, 6, 9, 12])

        # For both auto correlations and without them
        for auto_cor in (0, 1):

            def _create_ant_arrays():
                for remove_ants in removed_ants_per_time:
                    # Compute default antenna pairs
                    ant1, ant2 = np.triu_indices(na, auto_cor)

                    # Shuffle the antenna indices
                    idx = np.arange(ant1.size)
                    np.random.shuffle(idx)

                    ant1 = ant1[idx]
                    ant2 = ant2[idx]

                    # Remove any baselines containing flagged antenna
                    reduce_tuple = tuple(a != ra for a in (ant1, ant2)
                                         for ra in remove_ants)

                    keep = np.logical_and.reduce(reduce_tuple)
                    ant1 = ant1[keep]
                    ant2 = ant2[keep]

                    valid_ants = list(set(range(na)).difference(remove_ants))

                    yield valid_ants, remove_ants, ant1, ant2

            tup = zip(*list(_create_ant_arrays()))
            valid_ants, remove_ants, ant1, ant2 = tup

            bl_uvw = []

            # Create per-baseline UVW coordinates for each time chunk
            it = enumerate(zip(valid_ants, remove_ants, ant1, ant2))
            for t, (va, ra, a1, a2) in it:
                # Create random per-antenna UVW coordinates.
                # zeroing the first valid antenna
                ant_uvw = np.random.random(size=(na, 3)).astype(np.float64)
                ant_uvw[va[0], :] = 0
                # Create per-baseline UVW coordinates for this time chunk
                bl_uvw.append(ant_uvw[a1, :] - ant_uvw[a2, :])

            # Produced concatenated antenna and baseline uvw arrays
            time_chunks = np.array([a.size for a in ant1], dtype=ant1[0].dtype)
            cant1 = np.concatenate(ant1)
            cant2 = np.concatenate(ant2)
            cbl_uvw = np.concatenate(bl_uvw)

            # Now recover the per-antenna and per-baseline UVW coordinates
            # for the ntime chunks
            rant_uvw = antenna_uvw(cbl_uvw, cant1, cant2, time_chunks,
                                   nr_of_antenna=na, check_decomposition=True)


if __name__ == "__main__":
    unittest.main()



import string

# Template for the main header file
MAIN_HEADER_TEMPLATE = string.Template(
"""#ifndef ${main_header_guard}
#define ${main_header_guard}

// ${project} namespace start and stop defines
#define ${project_namespace_start} namespace ${project} {
#define ${project_namespace_stop} }

// ${snake_case} namespace start and stop defines
#define ${op_namespace_start} namespace ${snake_case} {
#define ${op_namespace_stop} }

${project_namespace_start}
${op_namespace_start}

// General definition of the ${opname} op, which will be specialised for CPUs and GPUs in
// ${cpp_header_file} and ${cuda_header_file} respectively, as well as float types (FT).
// Concrete template instantiations of this class should be provided in
// ${cpp_source_file} and ${cuda_source_file} respectively
template <typename Device, typename FT> class ${opname} {};

${op_namespace_stop}
${project_namespace_stop}

#endif // #ifndef ${main_header_guard}
""")





# Template for the c++ header file (CPU)
CPP_HEADER_TEMPLATE = string.Template(
"""#ifndef ${cpp_header_guard}
#define ${cpp_header_guard}

#include "${main_header_file}"

// Required in order for Eigen::ThreadPoolDevice to be an actual type
#define EIGEN_USE_THREADS

#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"

${project_namespace_start}
${op_namespace_start}

// For simpler partial specialisation
typedef Eigen::ThreadPoolDevice CPUDevice;

// Specialise the ${opname} op for CPUs
template <typename FT>
class ${opname}<CPUDevice, FT> : public tensorflow::OpKernel
{
public:
    explicit ${opname}(tensorflow::OpKernelConstruction * context) :
        tensorflow::OpKernel(context) {}

    void Compute(tensorflow::OpKernelContext * context) override
    {
        namespace tf = tensorflow;

        const tf::Tensor & in_input = context->input(0);

        // Allocate an output tensor
        tf::Tensor * output_ptr = nullptr;
        OP_REQUIRES_OK(context, context->allocate_output(
            0, in_input.shape(), &output_ptr));


        int N = in_input.dim_size(0);
        auto input = in_input.tensor<FT, 1>();
        auto output = output_ptr->tensor<FT, 1>();

        for(int i=0; i<N; ++i)
            { output(i) = input(i) + FT(1.0); }
    }
};

${op_namespace_stop}
${project_namespace_stop}

#endif // #ifndef ${cpp_header_guard}
""")





# Template for the c++ source file (CPU)
CPP_SOURCE_TEMPLATE = string.Template(
"""#include "${cpp_header_file}"

#include "tensorflow/core/framework/shape_inference.h"

${project_namespace_start}
${op_namespace_start}

using tensorflow::shape_inference::InferenceContext;
using tensorflow::shape_inference::ShapeHandle;
using tensorflow::shape_inference::DimensionHandle;
using tensorflow::Status;

auto shape_function = [](InferenceContext* c) {
    // Dummies for tests
    ShapeHandle input;
    DimensionHandle d;

    ShapeHandle in = c->input(0);

    // Assert that in has 1 dimension
    TF_RETURN_WITH_CONTEXT_IF_ERROR(c->WithRank(in, 1, &input),
        "in shape must be [N, ] but is " + c->DebugString(in));

    // Assert that in has a certain size
    // TF_RETURN_WITH_CONTEXT_IF_ERROR(c->WithValue(c->Dim(in, 0), N, &d),
    //     "in shape must be [N, ] but is " + c->DebugString(in));

    // Infer the shape of the output tensor,
    // in this case, the same shape as our input tensor
    ShapeHandle out = c->MakeShape({
        c->Dim(in, 0)
    });

    // Set the shape of the first output
    c->set_output(0, out);

    // printf("output shape %s\\n", c->DebugString(out).c_str());;

    return Status::OK();
};

// Register the ${opname} operator.
REGISTER_OP("${opname}")
    .Input("in: FT")
    .Output("out: FT")
    .Attr("FT: {double, float} = DT_FLOAT")
    .SetShapeFn(shape_function);

// Register a CPU kernel for ${opname} that handles floats
REGISTER_KERNEL_BUILDER(
    Name("${opname}")
    .TypeConstraint<float>("FT")
    .Device(tensorflow::DEVICE_CPU),
    ${opname}<CPUDevice, float>);

// Register a CPU kernel for ${opname} that handles doubles
REGISTER_KERNEL_BUILDER(
    Name("${opname}")
    .TypeConstraint<double>("FT")
    .Device(tensorflow::DEVICE_CPU),
    ${opname}<CPUDevice, double>);


${op_namespace_stop}
${project_namespace_stop}
""")





# Template for the cuda header file (GPU)
CUDA_HEADER_TEMPLATE = string.Template(
"""#if GOOGLE_CUDA

#ifndef ${cuda_header_guard}
#define ${cuda_header_guard}

#include "${main_header_file}"

// Required in order for Eigen::GpuDevice to be an actual type
#define EIGEN_USE_GPU

#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"

${project_namespace_start}
${op_namespace_start}

// For simpler partial specialisation
typedef Eigen::GpuDevice GPUDevice;

// LaunchTraits struct defining
// kernel block sizes for floats and doubles
template <typename FT> struct LaunchTraits {};

template <> struct LaunchTraits<float>
    { static constexpr int BLOCKDIMX = 1024; };

template <> struct LaunchTraits<double>
    { static constexpr int BLOCKDIMX = 1024; };

// CUDA kernel outline
template <typename FT>
__global__ void ${kernel_name}(const FT * input, FT * output, int N)
{
    // Shared memory usage unnecesssary, but demonstrates use of
    // constant Trait members to create kernel shared memory.
    using LTr = LaunchTraits<FT>;
    __shared__ FT buffer[LTr::BLOCKDIMX];

    int i = blockIdx.x*blockDim.x + threadIdx.x;

    if(i >= N)
        { return; }

    // Load in our input and add one to it
    buffer[threadIdx.x] = input[i];
    buffer[threadIdx.x] += FT(1.0);

    // Write to the outpu
    output[i] = buffer[threadIdx.x];
}

// Specialise the ${opname} op for GPUs
template <typename FT>
class ${opname}<GPUDevice, FT> : public tensorflow::OpKernel
{
public:
    explicit ${opname}(tensorflow::OpKernelConstruction * context) :
        tensorflow::OpKernel(context) {}

    void Compute(tensorflow::OpKernelContext * context) override
    {
        namespace tf = tensorflow;

        const tf::Tensor & in_input = context->input(0);

        int N = in_input.dim_size(0);

        // Allocate an output tensor
        tf::Tensor * output_ptr = nullptr;
        OP_REQUIRES_OK(context, context->allocate_output(
            0, in_input.shape(), &output_ptr));

        using LTr = LaunchTraits<FT>;

        // Set up our CUDA thread block and grid
        dim3 block(LTr::BLOCKDIMX);
        dim3 grid((N + block.x - 1)/block.x);

        // Get the GPU device
        const auto & device = context->eigen_device<GPUDevice>();

        // Get pointers to flattened tensor data buffers
        auto const input = in_input.flat<FT>().data();
        auto output = output_ptr->flat<FT>().data();

        // Call the ${kernel_name} CUDA kernel
        ${kernel_name}<<<grid, block, 0, device.stream()>>>(
            input, output, N);
    }
};

${op_namespace_stop}
${project_namespace_stop}

#endif // #ifndef ${cuda_header_guard}

#endif // #if GOOGLE_CUDA
""")





# Template for the cuda source file (GPU)
CUDA_SOURCE_TEMPLATE = string.Template(
"""#if GOOGLE_CUDA

#include "${cuda_header_file}"

${project_namespace_start}
${op_namespace_start}

// Register a GPU kernel for ${opname} that handles floats
REGISTER_KERNEL_BUILDER(
    Name("${opname}")
    .TypeConstraint<float>("FT")
    .Device(tensorflow::DEVICE_GPU),
    ${opname}<GPUDevice, float>);

// Register a GPU kernel for ${opname} that handles doubles
REGISTER_KERNEL_BUILDER(
    Name("${opname}")
    .TypeConstraint<double>("FT")
    .Device(tensorflow::DEVICE_GPU),
    ${opname}<GPUDevice, double>);

${op_namespace_stop}
${project_namespace_stop}

#endif // #if GOOGLE_CUDA
""")





# Template for the python test code
PYTHON_SOURCE_TEMPLATE = string.Template(
"""import os

import numpy as np
import tensorflow as tf

# Load the library containing the custom operation
from montblanc.impl.rime.tensorflow import load_tf_lib
rime = load_tf_lib()

# Register the shape function for the operation
from tensorflow.python.framework import common_shapes
from tensorflow.python.framework import ops
ops.RegisterShape("${opname}")(common_shapes.call_cpp_shape_fn)

# Create some input and wrap it in a tensorflow Variable
np_array = np.random.random(size=512*1024).astype(np.float32)
tf_array = tf.Variable(np_array)

# Pin the compute to the CPU
with tf.device('/cpu:0'):
    expr_cpu = ${module}.${snake_case}(tf_array)

# Pin the compute to the GPU
with tf.device('/gpu:0'):
    expr_gpu = ${module}.${snake_case}(tf_array)

init_op = tf.global_variables_initializer()

with tf.Session() as S:
    S.run(init_op)

    # Run our expressions on CPU and GPU
    result_cpu = S.run(expr_cpu)
    result_gpu = S.run(expr_gpu)

    # Check that 1.0 has been added to the input
    # and that CPU and GPU results agree
    assert np.allclose(result_cpu, np_array + 1.0)
    assert np.allclose(result_cpu, result_gpu)

""")


import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib


def brightness_numpy(stokes, pol_type):
    nsrc, ntime, nchan, _ = stokes.shape

    I = stokes[:, :, :, 0]  # noqa
    Q = stokes[:, :, :, 1]
    U = stokes[:, :, :, 2]
    V = stokes[:, :, :, 3]

    if pol_type == "linear":
        pass
    elif pol_type == "circular":
        Q, U, V = V, Q, U
    else:
        raise ValueError("Invalid pol_type '{}'".format(pol_type))

    CT = np.complex128 if stokes.dtype == np.float64 else np.complex64

    # Compute the brightness matrix
    B = np.empty(shape=(nsrc, ntime, nchan, 4), dtype=CT)
    B[:, :, :, 0] = I+Q
    B[:, :, :, 1] = U+V*1j
    B[:, :, :, 2] = U-V*1j
    B[:, :, :, 3] = I-Q

    return B


class TestBSqrt(unittest.TestCase):
    """ Tests the BSqrt operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                                if d.device_type == 'GPU']

    def test_b_sqrt(self):
        """ Test the BSqrt operator """
        # List of type constraint for testing this operator
        permutations = [[np.float32, np.complex64, 'linear', {'rtol': 1e-3}],
                        [np.float32, np.complex64, 'circular', {'rtol': 1e-3}],
                        [np.float64, np.complex128, 'linear', {}],
                        [np.float64, np.complex128, 'circular', {}]]

        # Run test with the type combinations above
        for FT, CT, pol_type, tols in permutations:
            self._impl_test_b_sqrt(FT, CT, pol_type, tols)

    def _impl_test_b_sqrt(self, FT, CT, pol_type, tols):
        """ Implementation of the BSqrt operator test """

        nsrc, ntime, na, nchan = 10, 50, 27, 32

        # Useful random floats functor
        def rf(*s):
            return np.random.random(size=s).astype(FT)

        def rc(*s):
            return (rf(*s) + 1j*rf(*s)).astype(CT)

        # Set up our numpy input arrays

        # Stokes parameters, should produce a positive definite matrix
        stokes = np.empty(shape=(nsrc, ntime, nchan, 4), dtype=FT)
        Q = stokes[:, :, :, 1] = rf(nsrc, ntime, nchan) - 0.5
        U = stokes[:, :, :, 2] = rf(nsrc, ntime, nchan) - 0.5
        V = stokes[:, :, :, 3] = rf(nsrc, ntime, nchan) - 0.5
        noise = rf(nsrc, ntime, nchan)*0.1
        # Need I^2 = Q^2 + U^2 + V^2 + noise^2
        stokes[:, :, :, 0] = np.sqrt(Q**2 + U**2 + V**2 + noise)

        # Choose random flux to invert
        mask = np.random.randint(0, 2, size=(nsrc, ntime, nchan)) == 1
        stokes[mask, 0] = -stokes[mask, 0]

        # Make the last matrix zero to test the positive semi-definite case
        stokes[-1, -1, -1, :] = 0

        # Argument list
        np_args = [stokes]
        # Argument string name list
        arg_names = ["stokes"]

        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.b_sqrt(*tf_args, CT=CT,
                                        polarisation_type=pol_type)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU bsqrt and invert flag
            cpu_bsqrt, cpu_invert = S.run(cpu_op)

            # Get our actual brightness matrices
            b = brightness_numpy(stokes, pol_type)
            b_2x2 = b.reshape(nsrc, ntime, nchan, 2, 2)
            b_sqrt_2x2 = cpu_bsqrt.reshape(nsrc, ntime, nchan, 2, 2)

            # Multiplying the square root matrix
            # by it's hermitian transpose
            square = np.einsum("...ij,...kj->...ik",
                               b_sqrt_2x2, b_sqrt_2x2.conj())

            # Apply any sign inversions
            square[:, :, :, :, :] *= cpu_invert[:, :, :, None, None]

            # And we should obtain the brightness matrix
            assert np.allclose(b_2x2, square, **tols)

            # Compare with GPU bsqrt and invert flag
            for gpu_bsqrt, gpu_invert in S.run(gpu_ops):
                self.assertTrue(np.all(cpu_invert == gpu_invert))

                # Compare cpu and gpu
                d = np.isclose(cpu_bsqrt, gpu_bsqrt, **tols)
                d = np.invert(d)
                p = np.nonzero(d)

                if p[0].size == 0:
                    continue

                import itertools
                it = (np.asarray(p).T, cpu_bsqrt[d], gpu_bsqrt[d])
                it = enumerate(zip(*it))

                msg = ["%s %s %s %s %s" % (i, idx, c, g, c-g)
                       for i, (idx, c, g) in it]

                self.fail("CPU/GPU bsqrt failed likely because the "
                          "last polarisation for each entry differs slightly "
                          "for FT=np.float32 and CT=np.complex64. "
                          "FT='%s' CT='%s'."
                          "Here are the values\n%s" % (FT, CT, '\n'.join(msg)))


if __name__ == "__main__":
    unittest.main()






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import unittest


def print_versions():
    import numpy
    import montblanc

    """
    Print the versions of software relied upon by montblanc.
    Inspired by numexpr testing suite.
    """
    print(('-=' * 38))
    print(('Python version:    %s' % sys.version))
    print(('Montblanc version: %s' % montblanc.__version__))
    print(("NumPy version:     %s" % numpy.__version__))

    if os.name == 'posix':
        (sysname, nodename, release, version, machine) = os.uname()
        print(('Platform:          %s-%s' % (sys.platform, machine)))

    print(('-=' * 38))

def suite():
    test_suite = unittest.TestSuite()
    niter = 1

    for n in range(niter):
        pass

    return test_suite

def test():
    print_versions()
    return unittest.TextTestRunner().run(suite())




import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib

class TestEBeam(unittest.TestCase):
    """ Tests the EBeam operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                                if d.device_type == 'GPU']

    def test_e_beam(self):
        """ Test the EBeam operator """
        # List of type constraint for testing this operator
        type_permutations = [[np.float32, np.complex64],
                             [np.float64, np.complex128]]

        # Run test with the type combinations above
        for FT, CT in type_permutations:
            self._impl_test_e_beam(FT, CT)

    def _impl_test_e_beam(self, FT, CT):
        """ Implementation of the EBeam operator test """

        nsrc, ntime, na, nchan = 20, 29, 14, 64
        beam_lw = beam_mh = beam_nud = 50

        # Useful random floats functor
        rf = lambda *s: np.random.random(size=s).astype(FT)
        rc = lambda *s: (rf(*s) + 1j*rf(*s)).astype(CT)

        # Set up our numpy input arrays
        lm = (rf(nsrc, 2) - 0.5) * 1e-1
        frequency = np.linspace(1e9, 2e9, nchan,dtype=FT)
        point_errors = (rf(ntime, na, nchan, 2) - 0.5) * 1e-2
        antenna_scaling = rf(na, nchan, 2)
        parallactic_angle = np.deg2rad(rf(ntime, na))
        parallactic_angle_sin = np.sin(parallactic_angle)
        parallactic_angle_cos = np.cos(parallactic_angle)
        beam_extents = FT([-0.9, -0.8, 1e9, 0.8, 0.9, 2e9])
        beam_freq_map = np.linspace(1e9, 2e9, beam_nud, dtype=FT, endpoint=True)
        e_beam = rc(beam_lw, beam_mh, beam_nud, 4)

        # Argument list
        np_args = [lm, frequency, point_errors, antenna_scaling,
                     parallactic_angle_sin, parallactic_angle_cos,
                     beam_extents, beam_freq_map, e_beam]
        # Argument string name list
        arg_names = ["lm", "frequency", "point_errors", "antenna_scaling",
                     "parallactic_angle_sin", "parallactic_angle_cos",
                     "beam_extents", "beam_freq_map", "e_beam"]

        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.e_beam(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU ejones
            cpu_ejones = S.run(cpu_op)

            # Check that most of them are non-zero
            nz_cpu = np.count_nonzero((cpu_ejones))
            self.assertTrue(nz_cpu > 0.8*cpu_ejones.size,
                "Less than 80% of the ejones turns are non-zero")

            # Compare with GPU ejones
            for gpu_ejones in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_ejones, gpu_ejones),
                    "This may fail due to discrepancies when rounding "
                    "floating point values near an integer value. "
                    "As the CPU and GPU may slightly differ, "
                    "values may be slightly below and integer on the CPU "
                    "and slghtly above on the GPU for instance. "
                    "It may be appropriate to ignore this assert check. ")

                proportion_acceptable = 1e-4
                d = np.invert(np.isclose(cpu_ejones, gpu_ejones))
                incorrect = d.sum()
                proportion_incorrect = incorrect / float(d.size)

                self.assertTrue(proportion_incorrect < proportion_acceptable,
                    "Proportion of incorrect E beam values {pi} "
                    "({i} out of {t}) is greater than the "
                    "accepted tolerance {pa}.".format(
                        pi=proportion_incorrect, i=incorrect,
                        t=d.size, pa=proportion_acceptable))

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import numpy as np

import montblanc
import montblanc.util as mbu

from montblanc.util import (
    random_float as rf,
    random_complex as rc)

def array_dict(name, shape, dtype, **kwargs):
    tags = kwargs.pop('tags', ())

    if isinstance(tags, str):
        tags = set(s.strip() for s in tags.split(","))
    else:
        tags = set(str(s).strip() for s in tags)

    D = {
        'name' : name,
        'shape' : shape,
        'dtype' : dtype,
        'tags' : tags,
    }

    D.update(kwargs)
    return D

def prop_dict(name,dtype,default):
    return {
        'name' : name,
        'dtype' : dtype,
        'default' : default,
        'setter' : True
    }

# Set up gaussian scaling parameters
# Derived from https://github.com/ska-sa/meqtrees-timba/blob/master/MeqNodes/src/PSVTensor.cc#L493
# and https://github.com/ska-sa/meqtrees-timba/blob/master/MeqNodes/src/PSVTensor.cc#L602
fwhm2int = 1.0/np.sqrt(np.log(256))

# List of properties
P = [
    prop_dict('sigma_sqrd', 'ft', 1.0),
    prop_dict('X2', 'ft', 0.0),
]

def default_base_ant_pairs(self, context):
    """ Compute base antenna pairs """
    k = 0 if context.cfg['auto_correlations'] == True else 1
    na = context.dim_global_size('na')
    gen = (i.astype(context.dtype) for i in np.triu_indices(na, k))

    # Cache np.triu_indices(na, k) as its likely that (na, k) will
    # stay constant much of the time. Assumption here is that this
    # method will be grafted onto a DefaultsSourceProvider with
    # the appropriate members.
    if self._is_cached:
        array_cache = self._chunk_cache['default_base_ant_pairs']
        key = (k, na)

        # Cache miss
        if key not in array_cache:
            array_cache[key] = tuple(gen)

        return array_cache[key]

    return tuple(gen)

def default_antenna1(self, context):
    """ Default antenna1 values """
    ant1, ant2 = default_base_ant_pairs(self, context)
    (tl, tu), (bl, bu) = context.dim_extents('ntime', 'nbl')
    ant1_result = np.empty(context.shape, context.dtype)
    ant1_result[:,:] = ant1[np.newaxis,bl:bu]
    return ant1_result

def default_antenna2(self, context):
    """ Default antenna2 values """
    ant1, ant2 = default_base_ant_pairs(self, context)
    (tl, tu), (bl, bu) = context.dim_extents('ntime', 'nbl')
    ant2_result = np.empty(context.shape, context.dtype)
    ant2_result[:,:] = ant2[np.newaxis,bl:bu]
    return ant2_result

def rand_uvw(self, context):
    distance = 10
    (tl, tu), (al, au) = context.dim_extents('ntime', 'na')
    ntime, na = tu - tl, au - al

    # Distribute the antenna in a circle configuration
    ant_angles = 2*np.pi*np.arange(al, au)

    # Angular difference between each antenna
    ant_angle_diff = 2*np.pi/context.dim_global_size('na')

    # Space the time offsets for each antenna
    time_angle = (ant_angle_diff *
        np.arange(tl, tu)/context.dim_global_size('ntime'))

    # Compute offsets per antenna and timestep
    time_ant_angles = ant_angles[np.newaxis,:]+time_angle[:,np.newaxis]

    A = np.empty(context.shape, context.dtype)
    U, V, W = A[:,:,0], A[:,:,1], A[:,:,2]
    U[:] = distance*np.cos(time_ant_angles)
    V[:] = distance*np.sin(time_ant_angles)
    W[:] = np.random.random(size=(ntime,na))*0.1

    # All antenna zero coordinates are set to (0,0,0)
    A[:,0,:] = 0

    return A

def identity_on_pols(self, context):
    """
    Returns [[1, 0], tiled up to other dimensions
             [0, 1]]
    """
    A = np.empty(context.shape, context.dtype)
    A[:,:,:] = [[[1,0,0,1]]]
    return A

def default_stokes(self, context):
    """
    Returns [[1, 0], tiled up to other dimensions
             [0, 0]]
    """
    A = np.empty(context.shape, context.dtype)
    A[:,:,:] = [[[1,0,0,0]]]
    return A

def rand_stokes(self, context):
    # Should be (nsrc, ntime, 4)
    A = np.empty(context.shape, context.dtype)
    I, Q, U, V = A[:,:,0], A[:,:,1], A[:,:,2], A[:,:,3]
    noise = rf(I.shape, context.dtype)*0.1
    Q[:] = rf(Q.shape, context.dtype) - 0.5
    U[:] = rf(U.shape, context.dtype) - 0.5
    V[:] = rf(V.shape, context.dtype) - 0.5
    I[:] = np.sqrt(Q**2 + U**2 + V**2 + noise)

    return A

def default_gaussian_shape(self, context):
    # Should be (3, ngsrc)
    A = np.empty(context.shape, context.dtype)

    if A.size != 0:
        A[:,:] = [[0],[0],[1]] # el, em, eR

    return A

def rand_gaussian_shape(self, context):
    # Should be (3, ngsrc)
    if np.product(context.shape) == 0:
        return np.empty(context.shape, context.dtype)

    return np.random.random(size=(context.shape)) # el, em, eR

def default_sersic_shape(self, context):
    # Should be (3, nssrc)
    A = np.empty(context.shape, context.dtype)

    if A.size != 0:
        A[:,:] = [[0],[0],[1]] # e1, e2, eS

    return A

def test_sersic_shape(self, context):
    # Should be (3, nssrc)
    assert len(context.shape) == 2 and context.shape[0] == 3

    A = np.empty(context.shape, context.dtype)

    if A.size != 0:
        # Random values seem to create v. large discrepancies
        # between the CPU and GPU versions. Go with
        # non-random data here, as per Marzia's original code
        A[:,:] = [[0],[0],[np.pi/648000]] # e1, e2, eS

    return A

_freq_low = 1e9
_freq_high = 2e9
_ref_freq = 1.5e9

RADIANS = "Radians"
HERTZ = "Hertz"
METERS = "Meters"
JANSKYS = "Janskys"
SECONDS = "Seconds"
DIMENSIONLESS = "Dimensionless"

LM_DESCRIPTION = ("(l,m) coordinates for {st} sources. "
            "Offset relative to the phase centre.")
STOKES_DESCRIPTION = ("(I,Q,U,V) Stokes parameters.")
ALPHA_DESCRIPTION = ("Power term describing the distribution of a source's flux "
            "over frequency. Distribution is calculated as (nu/nu_ref)^alpha "
            "where nu is frequency.")
REF_FREQ_DESCRIPTION = ("Reference frequency for {st} sources.")

# Tag Description
#
# input: arrays that must be input
# output: arrays that must be output
# temporary: arrays used to hold temporary results
# constant:

# List of arrays
A = [

    # Antenna Positions
    array_dict('antenna_position', ('na', 3), 'ft',
        default     = lambda s, c: np.zeros(c.shape, c.dtype),
        test        = lambda s, c: np.ones(c.shape, c.dtype),
        tags        = "input",
        description = "Antenna coordinates",
        units       = METERS),

    # Timesteps
    array_dict('time', ('ntime',), 'ft',
        default     = lambda s, c: np.linspace(0, 100,
                                    c.shape[0], dtype=c.dtype),
        test        = lambda s, c: np.linspace(0, 100,
                                    c.shape[0], dtype=c.dtype),
        tags        = "input",
        description = "Timesteps",
        units       = SECONDS),

    # Phase Centre
    array_dict('phase_centre', (2,), 'ft',
        default     = lambda s, c: np.array([0,0], dtype=c.dtype),
        test        = lambda s, c: np.array([0,0], dtype=c.dtype),
        tags        = "input",
        description = "Phase Centre",
        units       = RADIANS),

    # UVW Coordinates
    array_dict('uvw', ('ntime', 'na', 3), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = rand_uvw,
        tags    = "input, constant",
        description = "UVW antenna coordinates, normalised "
            "relative to a reference antenna.",
        units   = METERS),

    array_dict('antenna1', ('ntime', 'nbl'), np.int32,
        default = default_antenna1,
        test    = default_antenna1,
        tags    = "input",
        description = "Index of the first antenna "
            "of the baseline pair."),

    array_dict('antenna2', ('ntime', 'nbl'), np.int32,
        default = default_antenna2,
        test    = default_antenna2,
        tags    = "input",
        description = "Index of the second antenna "
            "of the baseline pair."),

    # Frequency
    array_dict('frequency', ('nchan',), 'ft',
        default = lambda s, c: np.linspace(_freq_low, _freq_high,
                                c.shape[0], dtype=c.dtype),
        test    = lambda s, c: np.linspace(_freq_low, _freq_high,
                                c.shape[0], dtype=c.dtype),
        tags    = "input",
        description = "Frequency. Frequencies from multiple bands "
            "are stacked on top of each other. ",
        units   = HERTZ),

    # Feed angles for each antenna
    array_dict('feed_angles', ('na',), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: rf(c.shape, c.dtype)*np.pi,
        tags    = "input, constant",
        description = "Feed angles for each antenna.",
        units   = RADIANS),


    # Parallactic angles at each timestep for each antenna
    array_dict('parallactic_angles', ('ntime', 'na'), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: rf(c.shape, c.dtype)*np.pi,
        tags    = "input, constant",
        description = "Parallactic angles for each antenna.",
        units   = RADIANS),

    # Holographic Beam

    # Pointing errors
    array_dict('pointing_errors', ('ntime','na','nchan', 2), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: (rf(c.shape, c.dtype)-0.5)*1e-2,
        tags    = "input, constant",
        description = "Pointing errors for each antenna. "
            "The components express an offset in the (l,m) plane.",
        units   = RADIANS),

    # Antenna scaling factors
    array_dict('antenna_scaling', ('na','nchan',2), 'ft',
        default = lambda s, c: np.ones(c.shape, c.dtype),
        test    = lambda s, c: rf(c.shape, c.dtype),
        tags    = "input, constant",
        description = "Antenna scaling factors for each antenna. "
            "The components express a scale in the (l,m) plane.",
        units   = DIMENSIONLESS),


    # Extents of the beam.
    # First 3 values are lower coordinate for (l, m, frequency)
    # while the last 3 are the upper coordinates
    array_dict('beam_extents', (6,), 'ft',
        default = lambda s, c: c.dtype([-1, -1, _freq_low, 1, 1, _freq_high]),
        test    = lambda s, c: c.dtype([-1, -1, _freq_low, 1, 1, _freq_high]),
        tags    = "input",
        description = "Extents of the holographic beam cube. "
            "[l_low, m_low, freq_low, l_high, m_high, freq_high].",
        units   = "[{r}, {r}, {h}, {r}, {r}, {h}]".format(r=RADIANS, h=HERTZ)),

    array_dict('beam_freq_map', ('beam_nud',), 'ft',
        default  = lambda s, c: np.linspace(_freq_low, _freq_high,
                                c.shape[0], endpoint=True, dtype=c.dtype),
        test     = lambda s, c: np.linspace(_freq_low, _freq_high,
                                c.shape[0], endpoint=True, dtype=c.dtype),
        tags    = "input",
        description = "A map describing the frequency associated with each "
            "slice of the frequency dimension of the holographic beam cube.",
        units   = HERTZ),

    # Beam cube
    array_dict('ebeam', ('beam_lw', 'beam_mh', 'beam_nud', 4), 'ct',
        default = identity_on_pols,
        test    = lambda s, c: rc(c.shape, c.dtype),
        tags    = "input, constant",
        description = "Holographic beam cube providing "
            "a discretised representation of the antenna beam pattern. "
            "Used to simulate the Direction Dependent Effects (DDE) "
            " or E term of the RIME."
            "Composed of a frequency stack of (l,m) images.",
        units   = DIMENSIONLESS),

    # Direction-Independent Effects
    array_dict('direction_independent_effects', ('ntime', 'na', 'nchan', 'npol'), 'ct',
        default = identity_on_pols,
        test    = lambda s, c: rc(c.shape, c.dtype),
        tags    = "input, constant",
        description = "Array providing the Direction Independent Effects (DIE) "
            "or G term of the RIME, term for each antenna.",
        units   = DIMENSIONLESS),

    # Point Source Definitions
    array_dict('point_lm', ('npsrc',2), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: (rf(c.shape, c.dtype)-0.5)*1e-1,
        tags    = "input, constant",
        description = LM_DESCRIPTION.format(st="point"),
        units   = RADIANS),
    array_dict('point_stokes', ('npsrc','ntime', 'nchan', 4), 'ft',
        default = default_stokes,
        test    = rand_stokes,
        tags    = "input, constant",
        description = STOKES_DESCRIPTION,
        units   = JANSKYS),

    # Gaussian Source Definitions
    array_dict('gaussian_lm', ('ngsrc',2), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: (rf(c.shape, c.dtype)-0.5)*1e-1,
        tags    = "input, constant",
        description = LM_DESCRIPTION.format(st="gaussian"),
        units   = RADIANS),
    array_dict('gaussian_stokes', ('ngsrc','ntime', 'nchan', 4), 'ft',
        default = default_stokes,
        test    = rand_stokes,
        tags    = "input, constant",
        description = STOKES_DESCRIPTION,
        units   = JANSKYS),
    array_dict('gaussian_shape', (3, 'ngsrc'), 'ft',
        default = default_gaussian_shape,
        test    = rand_gaussian_shape,
        tags    = "input, constant",
        description = "Parameters describing the shape of a gaussian source. "
            "(lproj, mproj, ratio) where lproj and mproj are the projections of "
            "the major and minor ellipse axes onto the l and m axes. "
            "Ratio is ratio of the minor/major axes.",
        units   = "({r}, {r}, {d})".format(r=RADIANS, d=DIMENSIONLESS)),

    # Sersic Source Definitions
    array_dict('sersic_lm', ('nssrc',2), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: (rf(c.shape, c.dtype)-0.5)*1e-1,
        tags    = "input, constant",
        description = LM_DESCRIPTION.format(st="sersic"),
        units   = "Radians"),
    array_dict('sersic_stokes', ('nssrc','ntime', 'nchan', 4), 'ft',
        default = default_stokes,
        test    = rand_stokes,
        tags    = "input, constant",
        description = STOKES_DESCRIPTION,
        units   = JANSKYS),
    array_dict('sersic_shape', (3, 'nssrc'), 'ft',
        default = default_sersic_shape,
        test    = test_sersic_shape,
        tags    = "input, constant",
        description = "Parameters describing the shape of a sersic source. "
            "(e1, e2, eR). Further information is required here.",
        units   = "({r}, {r}, {d})".format(r=RADIANS, d=DIMENSIONLESS)),

    # Observation Data

    # Visibility flagging array
    array_dict('flag', ('ntime', 'nbl', 'nchan', 'npol'), np.uint8,
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: np.random.random_integers(0, 1,
            size=c.shape).astype(np.uint8),
        tags    = "input, constant",
        description = "Indicates whether a visibility should be flagged when "
            "computing a Residual or Chi-Squared value.",
        unut    = DIMENSIONLESS),
    # Weight array
    array_dict('weight', ('ntime','nbl','nchan', 'npol'), 'ft',
        default = lambda s, c: np.ones(c.shape, c.dtype),
        test    = lambda s, c: rf(c.shape, c.dtype),
        tags    = "input, constant",
        description = "Weight applied to the difference of observed and model "
            "visibilities when computing a Chi-Squared value.",
        units   = DIMENSIONLESS),
    # Observed Visibilities
    array_dict('observed_vis', ('ntime','nbl','nchan', 'npol'), 'ct',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: rc(c.shape, c.dtype),
        tags    = "input, constant",
        description = "Observed visibilities, used to compute residuals and "
            "Chi-Squared values."),

    # Model Visibilities
    array_dict('model_vis', ('ntime','nbl','nchan', 'npol'), 'ct',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: rc(c.shape, c.dtype),
        tags    = ("input, output, constant"),
        description = "Model visibilities. In the context of input, these values "
            "will be added to the model visibilities computed by the RIME. "
            "This mechanism allows visibilities to be accumulated over different "
            "models, for example. However, they are zeroed by default."
            "In the context of output, these are the RIME model visibilities "),

    # Chi-squared
    array_dict('chi_squared', (1,), 'ft',
        default = lambda s, c: np.zeros(c.shape, c.dtype),
        test    = lambda s, c: np.zeros(c.shape, c.dtype),
        tags    = "output",
        description = "Chi-squared value associated with a tile "
            "of model and observed visibilities. These can be summed "
            "to produce a chi-squared for the entire problem."),

    # Result arrays
    array_dict('bsqrt', ('nsrc', 'ntime', 'nchan', 'npol'), 'ct',
        tags="temporary"),
    array_dict('cplx_phase', ('nsrc','ntime','na','nchan'), 'ct',
        tags="temporary"),
    array_dict('ejones', ('nsrc','ntime','na','nchan', 'npol'), 'ct',
        tags="temporary"),
    array_dict('ant_jones', ('nsrc','ntime','na','nchan', 'npol'), 'ct',
        tags="temporary"),
    array_dict('sgn_brightness', ('nsrc', 'ntime'), np.int8,
        tags="temporary"),
    array_dict('source_shape', ('nsrc', 'ntime', 'nbl', 'nchan'), 'ft',
        tags="temporary"),
    array_dict('chi_sqrd_result', ('ntime','nbl','nchan'), 'ft',
        tags="temporary"),
]



import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib

class TestParallacticAngleSinCos(unittest.TestCase):
    """ Tests the ParallacticAngleSinCos operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                                if d.device_type == 'GPU']

    def test_parallactic_angle_sin_cos(self):
        """ Test the ParallacticAngleSinCos operator """
        # List of type constraint for testing this operator
        type_permutations = [np.float32, np.float64]

        # Run test with the type combinations above
        for FT in type_permutations:
            self._impl_test_parallactic_angle_sin_cos(FT)

    def _impl_test_parallactic_angle_sin_cos(self, FT):
        """ Implementation of the ParallacticAngleSinCos operator test """

        # Create input variables
        ntime = 10
        na = 7

        parallactic_angle = np.random.random(size=[ntime, na]).astype(FT)


        # Argument list
        np_args = [parallactic_angle]
        # Argument string name list
        arg_names = ['parallactic_angle']
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.parallactic_angle_sin_cos(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU sincos
            cpu_pa_sin, cpu_pa_cos = S.run(cpu_op)

            # Compare with GPU sincos
            for gpu_pa_sin, gpu_pa_cos in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_pa_sin, gpu_pa_sin))
                self.assertTrue(np.allclose(cpu_pa_cos, gpu_pa_cos))

if __name__ == "__main__":
    unittest.main()


import ast

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

# builtin function whitelist
_BUILTIN_WHITELIST = frozenset(['slice'])
_missing = _BUILTIN_WHITELIST.difference(dir(__builtin__))
if len(_missing) > 0:
    raise ValueError("'%s' are not valid builtin functions.'" % list(_missing))


def parse_python_assigns(assign_str):
    """
    Parses a string, containing assign statements
    into a dictionary.

    .. code-block:: python

        h5 = katdal.open('123456789.h5')
        kwargs = parse_python_assigns("spw=3; scans=[1,2];"
                                      "targets='bpcal,radec';"
                                      "channels=slice(0,2048)")
        h5.select(**kwargs)

    Parameters
    ----------
    assign_str: str
        Assignment string. Should only contain assignment statements
        assigning python literals or builtin function calls, to variable names.
        Multiple assignment statements should be separated by semi-colons.

    Returns
    -------
    dict
        Dictionary { name: value } containing
        assignment results.
    """


    if not assign_str:
        return {}

    def _eval_value(stmt_value):
        # If the statement value is a call to a builtin, try evaluate it
        if isinstance(stmt_value, ast.Call):
            func_name = stmt_value.func.id

            if func_name not in _BUILTIN_WHITELIST:
                raise ValueError("Function '%s' in '%s' is not builtin. "
                                 "Available builtins: '%s'"
                                    % (func_name, assign_str, list(_BUILTIN_WHITELIST)))

            # Recursively pass arguments through this same function
            if stmt_value.args is not None:
                args = tuple(_eval_value(a) for a in stmt_value.args)
            else:
                args = ()

            # Recursively pass keyword arguments through this same function
            if stmt_value.keywords is not None:
                kwargs = {kw.arg : _eval_value(kw.value) for kw
                                          in stmt_value.keywords}
            else:
                kwargs = {}

            return getattr(__builtin__, func_name)(*args, **kwargs)
        # Try a literal eval
        else:
            return ast.literal_eval(stmt_value)

    # Variable dictionary
    variables = {}

    # Parse the assignment string
    stmts = ast.parse(assign_str, mode='single').body

    for i, stmt in enumerate(stmts):
        if not isinstance(stmt, ast.Assign):
            raise ValueError("Statement %d in '%s' is not a "
                             "variable assignment." % (i, assign_str))

        # Evaluate assignment lhs
        values = _eval_value(stmt.value)

        # "a = b = c" => targets 'a' and 'b' with 'c' as result
        for target in stmt.targets:
            # a = 2
            if isinstance(target, ast.Name):
                variables[target.id] = values

            # Tuple/List unpacking case
            # (a, b) = 2
            elif isinstance(target, (ast.Tuple, ast.List)):
                # Require all tuple/list elements to be variable names,
                # although anything else is probably a syntax error
                if not all(isinstance(e, ast.Name) for e in target.elts):
                    raise ValueError("Tuple unpacking in assignment %d "
                                     "in expression '%s' failed as not all "
                                     "tuple contents are variable names.")

                # Promote for zip and length checking
                if not isinstance(values, (tuple, list)):
                    elements = (values,)
                else:
                    elements = values

                if not len(target.elts) == len(elements):
                    raise ValueError("Unpacking '%s' into a tuple/list in "
                                     "assignment %d of expression '%s' failed. "
                                     "The number of tuple elements did not match "
                                     "the number of values."
                                        % (values, i, assign_str))

                # Unpack
                for variable, value in zip(target.elts, elements):
                    variables[variable.id] = value
            else:
                raise TypeError("'%s' types are not supported"
                                "as assignment targets." % type(target))


    return variables



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import montblanc

from .sink_provider import SinkProvider

class NullSinkProvider(SinkProvider):

    def name(self):
        return "Null"

    def model_vis(self, context):
        array_schema = context.array(context.name)
        slices = context.slice_index(*array_schema.shape)
        slice_str = ','.join('%s:%s' % (s.start, s.stop) for s in slices)
        montblanc.log.info("Received '{n}[{sl}]"
            .format(n=context.name, sl=slice_str))

    def chi_squared(self, context):
        array_schema = context.array(context.name)
        slices = context.slice_index(*array_schema.shape)
        slice_str = ','.join('%s:%s' % (s.start, s.stop) for s in slices)
        montblanc.log.info("Received '{n}[{sl}]"
            .format(n=context.name, sl=slice_str))

    def __str__(self):
        return self.__class__.__name__


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


from .context_help import context_help
from .hypercube_proxy_metaclass import HypercubeProxyMetaClass
import six

@six.add_metaclass(HypercubeProxyMetaClass)
class StartContext(object):
    """
    Start Context object passed to Providers.

    It provides information to the user implementing a data source
    about the extents of the data tile that should be provided.

    .. code-block:: python

        # uvw varies by time and baseline and has 3 coordinate components
        cube.register_array("uvw", ("ntime", "nbl", 3), np.float64)

        ...

        class CustomSourceProvider(SourceProvider):
            def start(self, start_context):
                # Query dimensions directly
                (lt, ut), (lb, ub) = context.dim_extents("ntime", "nbl")
                ...

    Public methods of a :py:class:`~hypercube.base_cube.HyperCube`
    are proxied on this object. Other useful information, such
    as the configuration, iteration space arguments are also
    present on this object.
    """
    __slots__ = ('_cube', '_cfg', '_iter_args')

    def __init__(self, cube, slvr_cfg, iter_args):
        self._cube = cube
        self._cfg = slvr_cfg
        self._iter_args = iter_args

    @property
    def cube(self):
        return self._cube

    @property
    def cfg(self):
        """
        Configuration
        """
        return self._cfg

    @cfg.setter
    def cfg(self, value):
        self._cfg = value

    @property
    def iter_args(self):
        """
        Iteration arguments that describe the tile sizes
        over which iteration is performed. In the following example,
        iteration is occuring in tiles of 100 Timesteps, 64 Channels
        and 50 Point Sources.

        .. code-block:: python

            context.iter_args == [("ntime", 100),
                    ("nchan", 64), ("npsrc", 50)]
        """
        return self._iter_args

    def help(self, display_cube=False):
        """
        Get help associated with this context

        Args
        -----
        display_cube: bool
            Add hypercube description to the output
        Returns
        -------
            str
                A help string associated with this context
        """
        return context_help(self, display_cube)